Package: textreuse
Type: Package
Title: Detect Text Reuse and Document Similarity
Version: 1.0.2
Date: 2026-07-24
Authors@R: c(
    person("Lincoln", "Mullen", role = "aut",
        comment = c(ORCID = "0000-0001-5103-6917")
    ),
    person("Yaoxiang", "Li", role = c("aut", "cre"),
        email = "liyaoxiang@outlook.com",
        comment = c(ORCID = "0000-0001-9200-1016")))
Description: Tools for measuring similarity among documents and detecting
    passages which have been reused. Implements shingled n-gram, skip n-gram,
    and other tokenizers; similarity/dissimilarity functions; pairwise
    comparisons; minhash and locality sensitive hashing algorithms; and a
    version of the Smith-Waterman local alignment algorithm suitable for
    natural language.
License: MIT + file LICENSE
URL: https://docs.ropensci.org/textreuse/,
        https://github.com/ropensci/textreuse
BugReports: https://github.com/ropensci/textreuse/issues
VignetteBuilder: knitr
Depends: R (>= 3.1.1)
Imports: assertthat (>= 0.1), digest (>= 0.6.8), dplyr (>= 1.1.0), NLP
        (>= 0.1.8), Matrix, Rcpp (>= 0.12.0), RcppProgress (>= 0.1),
        stringr (>= 1.0.0), tibble (>= 3.0.1), tidyr (>= 1.0.0)
Suggests: testthat (>= 0.11.0), knitr (>= 1.11), rmarkdown (>= 0.8),
        covr
LinkingTo: BH, Rcpp, RcppProgress
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-07-24 22:50:27 UTC; Li
Author: Lincoln Mullen [aut] (ORCID: <https://orcid.org/0000-0001-5103-6917>),
  Yaoxiang Li [aut, cre] (ORCID: <https://orcid.org/0000-0001-9200-1016>)
Maintainer: Yaoxiang Li <liyaoxiang@outlook.com>
Repository: CRAN
Date/Publication: 2026-07-25 17:10:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-07-25 19:53:43 UTC; unix
Archs: textreuse.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("[",TextReuseCorpus)
S3method("[[",TextReuseCorpus)
S3method("content<-",TextReuseTextDocument)
S3method("hashes<-",TextReuseTextDocument)
S3method("meta<-",TextReuseCorpus)
S3method("meta<-",TextReuseTextDocument)
S3method("minhashes<-",TextReuseTextDocument)
S3method("names<-",TextReuseCorpus)
S3method("tokens<-",TextReuseTextDocument)
S3method(align_local,TextReuseTextDocument)
S3method(align_local,default)
S3method(as.character,TextReuseTextDocument)
S3method(as.matrix,textreuse_candidates)
S3method(content,TextReuseTextDocument)
S3method(count_matches,TextReuseTextDocument)
S3method(count_matches,default)
S3method(hashes,TextReuseCorpus)
S3method(hashes,TextReuseTextDocument)
S3method(jaccard_bag_similarity,TextReuseTextDocument)
S3method(jaccard_bag_similarity,default)
S3method(jaccard_dissimilarity,default)
S3method(jaccard_similarity,TextReuseTextDocument)
S3method(jaccard_similarity,default)
S3method(length,TextReuseCorpus)
S3method(lsh,TextReuseCorpus)
S3method(lsh,TextReuseTextDocument)
S3method(matching_tokens,TextReuseTextDocument)
S3method(matching_tokens,default)
S3method(meta,TextReuseCorpus)
S3method(meta,TextReuseTextDocument)
S3method(minhashes,TextReuseCorpus)
S3method(minhashes,TextReuseTextDocument)
S3method(names,TextReuseCorpus)
S3method(print,TextReuseCorpus)
S3method(print,TextReuseTextDocument)
S3method(print,textreuse_alignment)
S3method(ratio_of_matches,TextReuseTextDocument)
S3method(ratio_of_matches,default)
S3method(rehash,TextReuseCorpus)
S3method(rehash,TextReuseTextDocument)
S3method(tokenize,TextReuseCorpus)
S3method(tokenize,TextReuseTextDocument)
S3method(tokens,TextReuseCorpus)
S3method(tokens,TextReuseTextDocument)
S3method(wordcount,TextDocument)
S3method(wordcount,TextReuseCorpus)
S3method(wordcount,default)
export("content<-")
export("hashes<-")
export("meta<-")
export("minhashes<-")
export("tokens<-")
export(TextReuseCorpus)
export(TextReuseTextDocument)
export(align_local)
export(as_sparse_matrix)
export(content)
export(count_matches)
export(filenames)
export(has_content)
export(has_hashes)
export(has_minhashes)
export(has_tokens)
export(hash_string)
export(hashes)
export(is.TextReuseCorpus)
export(is.TextReuseTextDocument)
export(jaccard_bag_similarity)
export(jaccard_dissimilarity)
export(jaccard_similarity)
export(lsh)
export(lsh_add)
export(lsh_candidates)
export(lsh_compare)
export(lsh_probability)
export(lsh_query)
export(lsh_subset)
export(lsh_threshold)
export(matching_tokens)
export(meta)
export(minhash_generator)
export(minhashes)
export(pairwise_candidates)
export(pairwise_compare)
export(ratio_of_matches)
export(rehash)
export(skipped)
export(token_index)
export(token_index_candidates)
export(tokenize)
export(tokenize_ngrams)
export(tokenize_sentences)
export(tokenize_skip_ngrams)
export(tokenize_words)
export(tokens)
export(wordcount)
import(RcppProgress)
import(assertthat)
import(stringr)
importFrom(NLP,"content<-")
importFrom(NLP,"meta<-")
importFrom(NLP,content)
importFrom(NLP,meta)
importFrom(Rcpp,sourceCpp)
importFrom(utils,getTxtProgressBar)
importFrom(utils,setTxtProgressBar)
importFrom(utils,txtProgressBar)
useDynLib(textreuse, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# textreuse

[![CRAN\_Status\_Badge](https://www.r-pkg.org/badges/version/textreuse)](https://cran.r-project.org/package=textreuse)
[![CRAN\_Downloads](https://cranlogs.r-pkg.org/badges/grand-total/textreuse)](https://cran.r-project.org/package=textreuse)
[![Coverage
Status](https://img.shields.io/codecov/c/github/ropensci/textreuse/master.svg)](https://app.codecov.io/github/ropensci/textreuse?branch=master)
[![rOpenSci
badge](https://badges.ropensci.org/20_status.svg)](https://github.com/ropensci/software-review/issues/20)

## Overview

This [R](https://www.r-project.org/) package provides a set of functions
for measuring similarity among documents and detecting passages which
have been reused. It implements shingled n-gram, skip n-gram, and other
tokenizers; similarity/dissimilarity functions; pairwise comparisons;
minhash and locality sensitive hashing algorithms; and a version of the
Smith-Waterman local alignment algorithm suitable for natural language.
It is broadly useful for, for example, detecting duplicate documents in
a corpus prior to text analysis, or for identifying borrowed passages
between texts. The classes provided by this package follow the model of
other natural language processing packages for R, especially the
[NLP](https://cran.r-project.org/package=NLP) and
[tm](https://cran.r-project.org/package=tm) packages. (However, this
package has no dependency on Java, which should make it easier to
install.)

### Citation

If you use this package for scholarly research, I would appreciate a
citation.

    citation("textreuse")
    #> To cite package 'textreuse' in publications use:
    #> 
    #>   Mullen L, Li Y (2026). _textreuse: Detect Text Reuse and Document
    #>   Similarity_. R package version 1.0.1,
    #>   https://github.com/ropensci/textreuse,
    #>   <https://docs.ropensci.org/textreuse/>.
    #> 
    #> A BibTeX entry for LaTeX users is
    #> 
    #>   @Manual{,
    #>     title = {textreuse: Detect Text Reuse and Document Similarity},
    #>     author = {Lincoln Mullen and Yaoxiang Li},
    #>     year = {2026},
    #>     note = {R package version 1.0.1, 
    #> https://github.com/ropensci/textreuse},
    #>     url = {https://docs.ropensci.org/textreuse/},
    #>   }

## Installation

To install this package from CRAN:

    install.packages("textreuse")

To install the development version from GitHub, use
[devtools](https://github.com/r-lib/devtools).

    # install.packages("devtools")
    devtools::install_github("ropensci/textreuse", build_vignettes = TRUE)

## Examples

There are three main approaches that one may take when using this
package: pairwise comparisons, minhashing/locality sensitive hashing,
and extracting matching passages through text alignment.

See the [introductory
vignette](https://docs.ropensci.org/textreuse/articles/textreuse-introduction.html)
for a description of the classes provided by this package.

    vignette("textreuse-introduction", package = "textreuse")

### Pairwise comparisons

In this example we will load a tiny corpus of three documents. These
documents are drawn from Kellen Funk’s
[research](https://kellenfunk.org/field-code/) into the propagation of
legal codes of civil procedure in the nineteenth-century United States.

    library(textreuse)
    dir <- system.file("extdata/legal", package = "textreuse")
    corpus <- TextReuseCorpus(dir = dir, meta = list(title = "Civil procedure"),
                              tokenizer = tokenize_ngrams, n = 7)

We have loaded the three documents into a corpus, which involves
tokenizing the text and hashing the tokens. We can inspect the corpus as
a whole or the individual documents that make it up.

    corpus
    #> TextReuseCorpus
    #> Number of documents: 3 
    #> hash_func : hash_string 
    #> title : Civil procedure 
    #> tokenizer : tokenize_ngrams
    names(corpus)
    #> [1] "ca1851-match"   "ca1851-nomatch" "ny1850-match"
    corpus[["ca1851-match"]]
    #> TextReuseTextDocument
    #> file : C:/Users/Bach/AppData/Local/R/win-library/4.4/textreuse/extdata/legal/ca1851-match.txt 
    #> hash_func : hash_string 
    #> id : ca1851-match 
    #> minhash_func : 
    #> tokenizer : tokenize_ngrams 
    #> content : § 4. Every action shall be prosecuted in the name of the real party
    #> in interest, except as otherwise provided in this Act.
    #> 
    #> § 5. In the case of an assignment of a thing in action, the action by
    #> the as

Now we can compare each of the documents to one another. The
`pairwise_compare()` function applies a comparison function (in this
case, `jaccard_similarity()`) to every pair of documents. The result is
a matrix of scores. As we would expect, some documents are similar and
others are not.

    comparisons <- pairwise_compare(corpus, jaccard_similarity)
    comparisons
    #>                ca1851-match ca1851-nomatch ny1850-match
    #> ca1851-match             NA              0    0.3842549
    #> ca1851-nomatch           NA             NA    0.0000000
    #> ny1850-match             NA             NA           NA

We can convert that matrix to a data frame of pairs and scores if we
prefer.

    pairwise_candidates(comparisons)
    #> # A tibble: 3 × 3
    #>   a              b              score
    #> * <chr>          <chr>          <dbl>
    #> 1 ca1851-match   ca1851-nomatch 0    
    #> 2 ca1851-match   ny1850-match   0.384
    #> 3 ca1851-nomatch ny1850-match   0

See the [pairwise
vignette](https://docs.ropensci.org/textreuse/articles/textreuse-pairwise.html)
for a fuller description.

    vignette("textreuse-pairwise", package = "textreuse")

### Minhashing and locality sensitive hashing

Pairwise comparisons can be very time-consuming because they grow
geometrically with the size of the corpus. (A corpus with 10 documents
would require at least 45 comparisons; a corpus with 100 documents would
require 4,950 comparisons; a corpus with 1,000 documents would require
499,500 comparisons.) That’s why this package implements the minhash and
locality sensitive hashing algorithms, which can detect candidate pairs
much faster than pairwise comparisons in corpora of any significant
size.

For this example we will load a small corpus of ten documents published
by the American Tract Society. We will also create a minhash function,
which represents an entire document (regardless of length) by a fixed
number of integer hashes. When we create the corpus, the documents will
each have a minhash signature.

    dir <- system.file("extdata/ats", package = "textreuse")
    minhash <- minhash_generator(200, seed = 235)
    ats <- TextReuseCorpus(dir = dir,
                           tokenizer = tokenize_ngrams, n = 5,
                           minhash_func = minhash)

Now we can calculate potential matches, extract the candidates, and
apply a comparison function to just those candidates.

    buckets <- lsh(ats, bands = 50, progress = FALSE)
    candidates <- lsh_candidates(buckets)
    scores <- lsh_compare(candidates, ats, jaccard_similarity, progress = FALSE)
    scores
    #> # A tibble: 1 × 3
    #>   a              b                      score
    #>   <chr>          <chr>                  <dbl>
    #> 1 remember00palm remembermeorholy00palm 0.701

For details, see the [minhash
vignette](https://docs.ropensci.org/textreuse/articles/textreuse-minhash.html).

    vignette("textreuse-minhash", package = "textreuse")

### Text alignment

We can also extract the optimal alignment between two documents with a
version of the
[Smith-Waterman](https://en.wikipedia.org/wiki/Smith-Waterman_algorithm)
algorithm, used for protein sequence alignment, adapted for natural
language. The longest matching substring according to scoring values
will be extracted, and variations in the alignment will be marked.

    a <- "'How do I know', she asked, 'if this is a good match?'"
    b <- "'This is a match', he replied."
    align_local(a, b)
    #> TextReuse alignment
    #> Alignment score: 7 
    #> Document A:
    #> this is a good match
    #> 
    #> Document B:
    #> This is a #### match

For details, see the [text alignment
vignette](https://docs.ropensci.org/textreuse/articles/textreuse-alignment.html).

    vignette("textreuse-alignment", package = "textreuse")

### Parallel processing

Loading the corpus and creating tokens benefit from using multiple
cores, if available. (This works only on non-Windows machines.) To use
multiple cores, set `options("mc.cores" = 4L)`, where the number is how
many cores you wish to use.

### Contributing and acknowledgments

Please note that this project is released with a [Contributor Code of
Conduct](https://github.com/ropensci/textreuse/blob/master/CONDUCT.md).
By participating in this project you agree to abide by its terms.

Thanks to [Noam Ross](https://www.noamross.net/) for his thorough [peer
review](https://github.com/ropensci/software-review/issues/20) of this
package for [rOpenSci](https://ropensci.org/).

------------------------------------------------------------------------

[![rOpenSCi
logo](https://ropensci.org/public_images/github_footer.png)](https://ropensci.org)

## ----eval=FALSE---------------------------------------------------------------
# vignette("textreuse-pairwise", package = "textreuse")
# vignette("textreuse-minhash", package = "textreuse")
# vignette("textreuse-alignment", package = "textreuse")

## -----------------------------------------------------------------------------
library(textreuse)
file <- system.file("extdata/ats/remember00palm.txt", 
                    package = "textreuse")
doc <- TextReuseTextDocument(file = file, meta = list("publisher" = "ATS"),
                             tokenizer = tokenize_ngrams, n = 5,
                             keep_tokens = TRUE)
doc

## -----------------------------------------------------------------------------
meta(doc)
meta(doc, "id")
meta(doc, "date") <- 1865
head(tokens(doc))
head(hashes(doc))
wordcount(doc)

## -----------------------------------------------------------------------------
dir <- system.file("extdata/ats", package = "textreuse")
corpus <- TextReuseCorpus(dir = dir, tokenizer = tokenize_ngrams, n = 5,
                          progress = FALSE)
corpus

## -----------------------------------------------------------------------------
names(corpus)
corpus[["remember00palm"]]
corpus[c("calltounconv00baxt", "lifeofrevrichard00baxt")]

## -----------------------------------------------------------------------------
wordcount(corpus)

## -----------------------------------------------------------------------------
text <- "How many roads must a man walk down\nBefore you'll call him a man?"

tokenize_words(text)
tokenize_sentences(text)
tokenize_ngrams(text, n = 3)
tokenize_skip_ngrams(text, n = 3, k = 2)

## -----------------------------------------------------------------------------
poem <- "Roses are red\nViolets are blue\nI like using R\nAnd you should too"
cat(poem)

tokenize_lines <- function(string) {
  stringr::str_split(string, "\n+")[[1]]
}

tokenize_lines(poem)

## -----------------------------------------------------------------------------
hash_string(tokenize_words(text))

## -----------------------------------------------------------------------------
a <- tokenize_words(paste("How does it feel, how does it feel?",
                          "To be without a home",
                          "Like a complete unknown, like a rolling stone"))
b <- tokenize_words(paste("How does it feel, how does it feel?",
                          "To be on your own, with no direction home",
                          "A complete unknown, like a rolling stone"))

jaccard_similarity(a, b)
jaccard_dissimilarity(a, b)
jaccard_bag_similarity(a, b)
ratio_of_matches(a, b)

## ----eval = FALSE-------------------------------------------------------------
# options("mc.cores" = 4L)

## -----------------------------------------------------------------------------
shakespeare <- paste(
  "Haste still pays haste, and leisure answers leisure;",
  "Like doth quit like, and MEASURE still FOR MEASURE.",
  "Then, Angelo, thy fault's thus manifested;",
  "Which, though thou wouldst deny, denies thee vantage.",
  "We do condemn thee to the very block",
  "Where Claudio stoop'd to death, and with like haste.",
  "Away with him!")
critic <- paste(
  "The play comes to its culmination where Duke Vincentio, quoting from",
  "the words of the Sermon on the Mount, says,",
  "'Haste still goes very quickly , and leisure answers leisure;",
  "Like doth cancel like, and measure still for measure.'",
  "These titular words sum up the meaning of the play.")

## -----------------------------------------------------------------------------
library(textreuse)
align_local(shakespeare, critic)

#' Convert candidates data frames to other formats
#'
#' These functions convert a \code{textreuse_candidates} object to dense or
#' sparse matrices.
#'
#' @param x An object of class \code{\link[=lsh_compare]{textreuse_candidates}}.
#' @param ... Additional arguments.
#'
#' @return A similarity matrix with row and column names containing document IDs.
#'
#' @export
#' @method as.matrix textreuse_candidates
as.matrix.textreuse_candidates <- function(x, ...) {

  docs <- candidate_doc_ids(x)
  n <- length(docs)
  m <- matrix(0, n, n)
  rownames(m) <- docs
  colnames(m) <- docs
  diag(m) <- 1.0

  for (r in seq_len(nrow(x))) {
    a <- x$a[r]
    b <- x$b[r]
    score <- x$score[r]
    m[a, b] <- score
    m[b, a] <- score
  }

  m

}

#' @rdname as.matrix.textreuse_candidates
#' @export
as_sparse_matrix <- function(x) {
  assert_that(is_candidates_df(x))

  docs <- candidate_doc_ids(x)
  n <- length(docs)
  doc_ids <- stats::setNames(seq_along(docs), docs)

  rows <- seq_len(n)
  cols <- seq_len(n)
  values <- rep(1.0, n)

  if (nrow(x) > 0) {
    a <- unname(doc_ids[x$a])
    b <- unname(doc_ids[x$b])
    rows <- c(rows, a, b)
    cols <- c(cols, b, a)
    values <- c(values, x$score, x$score)
  }

  Matrix::sparseMatrix(i = rows, j = cols, x = values, dims = c(n, n),
                       use.last.ij = TRUE,
                       dimnames = list(docs, docs))
}

candidate_doc_ids <- function(x) {
  all_doc_ids <- attr(x, "all-doc-ids")
  if (is.null(all_doc_ids)) {
    all_doc_ids <- c(x$a, x$b)
  }
  sort(unique(all_doc_ids))
}

context("Candidate pairs")
dir <- system.file("extdata/legal", package = "textreuse")
corpus <- TextReuseCorpus(dir = dir)
m <- pairwise_compare(corpus, ratio_of_matches, directional = TRUE)
pairs <- pairwise_candidates(m)

test_that("returns a data frame with correct properties", {
  expect_equal(nrow(pairs), 6)
  expect_is(pairs, "data.frame")
  expect_is(pairs, "tbl_df")
  expect_is(pairs, "textreuse_candidates")
  expect_named(pairs, c("a", "b", "score"))
})

test_that("can be converted to matrix", {
  m <- as.matrix(pairs)
  expect_is(m, "matrix")
  expect_equal(names(corpus), colnames(m))
})

test_that("matrix conversion can keep all document ids", {
  pairs_with_ids <- pairs[1, ]
  attr(pairs_with_ids, "all-doc-ids") <- names(corpus)
  m <- as.matrix(pairs_with_ids)
  expect_equal(names(corpus), colnames(m))
  expect_equal(names(corpus), rownames(m))
})

test_that("can be converted to sparse matrix", {
  m <- as_sparse_matrix(pairs)
  expect_is(m, "dgCMatrix")
  expect_equal(names(corpus), colnames(m))
  expect_equal(as.matrix(m), as.matrix(pairs))
})

#' Filenames from paths
#'
#' This function takes a character vector of paths and returns just the file
#' name, by default without the extension. A \code{\link{TextReuseCorpus}} uses
#' the paths to the files in the corpus as the names of the list. This function
#' is intended to turn those paths into more manageable identifiers.
#'
#' @param paths A character vector of paths.
#' @param extension Should the file extension be preserved?
#' @seealso \code{\link{basename}}
#' @examples
#' paths <- c("corpus/one.txt", "corpus/two.md", "corpus/three.text")
#' filenames(paths)
#' filenames(paths, extension = TRUE)
#' @export
filenames <- function(paths, extension = FALSE) {
  assert_that(is.character(paths))
  f <- basename(paths)
  if (extension)
    return(f)
  else
    str_replace(f, "\\.[:alpha:]{1,}$", "")
}

context("Filenames")

paths <- c("corpus/one.txt", "deep/corpus/two.R", "~/home/three.markdown",
           "/corpus/four.md", "../corpus/five.text")
test_that("properly returns filenames", {
  expect_equal(filenames(paths), c("one", "two", "three", "four", "five"))
  expect_equal(filenames(paths, extension = TRUE), basename(paths))
})

#' Compare candidates identified by LSH
#'
#' The \code{\link{lsh_candidates}} only identifies potential matches, but
#' cannot estimate the actual similarity of the documents. This function takes a
#' data frame returned by \code{\link{lsh_candidates}} and applies a comparison
#' function to each of the documents in a corpus, thereby calculating the
#' document similarity score. Note that since your corpus will have minhash
#' signatures rather than hashes for the tokens itself, you will probably wish
#' to use \code{\link{tokenize}} to calculate new hashes. This can be done for
#' just the potentially similar documents. See the package vignettes for
#' details.
#'
#' @param candidates A data frame returned by \code{\link{lsh_candidates}}.
#' @param corpus The same \code{\link{TextReuseCorpus}} corpus which was used to generate the candidates.
#' @param f A comparison function such as \code{\link{jaccard_similarity}}.
#' @param progress Display a progress bar while comparing documents. Progress
#'   bars are disabled when using parallel processing.
#' @return A data frame with values calculated for \code{score}.
#' @examples
#' dir <- system.file("extdata/legal", package = "textreuse")
#' minhash <- minhash_generator(200, seed = 234)
#' corpus <- TextReuseCorpus(dir = dir,
#'                           tokenizer = tokenize_ngrams, n = 5,
#'                           minhash_func = minhash)
#' buckets <- lsh(corpus, bands = 50)
#' candidates <- lsh_candidates(buckets)
#' lsh_compare(candidates, corpus, jaccard_similarity)
#' @export
lsh_compare <- function(candidates, corpus, f, progress = interactive()) {
  assert_that(is_candidates_df(candidates),
              is.function(f),
              is.TextReuseCorpus(corpus))

  rows_to_score <- which(is.na(candidates$score))
  num_rows <- length(rows_to_score)
  use_parallel <- using_parallel()

  if (num_rows == 0) {
    attr(candidates, "all-doc-ids") <- names(corpus)
    return(candidates)
  }

  if (progress) {
    message("Making ", prettyNum(num_rows, big.mark = ","),
            " comparisons.")
    if (!use_parallel) {
      pb <- txtProgressBar(min = 0, max = num_rows, style = 3)
    }
  }

  apply_fun <- get_apply_function()
  scores <- apply_fun(seq_along(rows_to_score), function(j) {
    i <- rows_to_score[j]
    a <- candidates$a[i]
    b <- candidates$b[i]
    score <- f(corpus[[a]], corpus[[b]])
    if (progress && !use_parallel) setTxtProgressBar(pb, j)
    score
  })

  candidates$score[rows_to_score] <- unlist(scores, use.names = FALSE)

  if (progress && !use_parallel) close(pb)

  attr(candidates, "all-doc-ids") <- names(corpus)

  candidates
}

#' Measure similarity/dissimilarity in documents
#'
#' A set of functions which take two sets or bag of words and measure their
#' similarity or dissimilarity.
#'
#' @details The functions \code{jaccard_similarity} and
#'   \code{jaccard_dissimilarity} provide the Jaccard measures of similarity or
#'   dissimilarity for two sets. The coefficients will be numbers between
#'   \code{0} and \code{1}. For the similarity coefficient, the higher the
#'   number the more similar the two sets are. When applied to two documents of
#'   class \code{\link{TextReuseTextDocument}}, the hashes in those documents
#'   are compared. But this function can be passed objects of any class accepted
#'   by the set functions in base R. So it is possible, for instance, to pass
#'   this function two character vectors comprised of word, line, sentence, or
#'   paragraph tokens, or those character vectors hashed as integers.
#'
#'   The Jaccard similarity coeffecient is defined as follows:
#'
#'   \deqn{J(A, B) = \frac{ | A \cap B | }{ | A \cup B | }}{ length(intersect(a,
#'   b)) / length(union(a, b))}
#'
#'   The Jaccard dissimilarity is simply
#'
#'   \deqn{1 - J(A, B)}
#'
#'   The function \code{jaccard_bag_similarity} treats \code{a} and \code{b} as
#'   bags rather than sets, so that the result is a fraction where the numerator
#'   is the sum of each matching element counted the minimum number of times it
#'   appears in each bag, and the denominator is the sum of the lengths of both
#'   bags. The maximum value for the Jaccard bag similarity is \code{0.5}.
#'
#'   The function \code{ratio_of_matches} finds the ratio between the number of
#'   items in \code{b} that are also in \code{a} and the total number of items
#'   in \code{b}. Note that this similarity measure is directional: it measures
#'   how much \code{b} borrows from \code{a}, but says nothing about how much of
#'   \code{a} borrows from \code{b}.
#'
#'   The function \code{count_matches} returns the numerator used by
#'   \code{ratio_of_matches}: the number of items in \code{b} also found in
#'   \code{a}. The function \code{matching_tokens} returns those matching items
#'   from \code{b}, preserving their order and duplicates.
#'
#' @param a The first set (or bag) to be compared. The origin bag for
#'   directional comparisons.
#' @param b The second set (or bag) to be compared. The destination bag for
#'   directional comparisons.
#'
#' @examples
#' jaccard_similarity(1:6, 3:10)
#' jaccard_dissimilarity(1:6, 3:10)
#'
#' a <- c("a", "a", "a", "b")
#' b <- c("a", "a", "b", "b", "c")
#' jaccard_similarity(a, b)
#' jaccard_bag_similarity(a, b)
#' ratio_of_matches(a, b)
#' ratio_of_matches(b, a)
#' count_matches(a, b)
#' matching_tokens(a, b)
#'
#' ny         <- system.file("extdata/legal/ny1850-match.txt", package = "textreuse")
#' ca_match   <- system.file("extdata/legal/ca1851-match.txt", package = "textreuse")
#' ca_nomatch <- system.file("extdata/legal/ca1851-nomatch.txt", package = "textreuse")
#'
#' ny         <- TextReuseTextDocument(file = ny,
#'                                     meta = list(id = "ny"))
#' ca_match   <- TextReuseTextDocument(file = ca_match,
#'                                     meta = list(id = "ca_match"))
#' ca_nomatch <- TextReuseTextDocument(file = ca_nomatch,
#'                                     meta = list(id = "ca_nomatch"))
#'
#' # These two should have higher similarity scores
#' jaccard_similarity(ny, ca_match)
#' ratio_of_matches(ny, ca_match)
#'
#' # These two should have lower similarity scores
#' jaccard_similarity(ny, ca_nomatch)
#' ratio_of_matches(ny, ca_nomatch)
#'
#' @references Jure Leskovec, Anand Rajaraman, and Jeff Ullman,
#'   \emph{Mining of Massive Datasets} (Cambridge University Press, 2011).
#' @name similarity-functions
NULL

#' @rdname similarity-functions
#' @export
jaccard_similarity <- function(a, b) UseMethod("jaccard_similarity")

#' @export
jaccard_similarity.default <- function(a, b) {
  assert_that(all(class(a) == class(b)))
  length(intersect(a, b)) / length(union(a, b))
}

#' @export
jaccard_similarity.TextReuseTextDocument <- function(a, b) {
  assert_that(all(class(a) == class(b)))
  jaccard_similarity(a$hashes, b$hashes)
}

#' @rdname similarity-functions
#' @export
jaccard_dissimilarity <- function(a, b) UseMethod("jaccard_dissimilarity")

#' @export
jaccard_dissimilarity.default <- function(a, b) {
  1 - jaccard_similarity(a, b)
}

#' @rdname similarity-functions
#' @export
jaccard_bag_similarity <- function(a, b) UseMethod("jaccard_bag_similarity")

#' @export
jaccard_bag_similarity.default <- function(a, b) {
  matches <- intersect(a, b)
  counts <- vapply(matches, function(x) min(sum(x == a), sum(x == b)),
                   integer(1), USE.NAMES = FALSE)
  denominator <- length(a) + length(b)
  sum(counts) / denominator
}

#' @export
jaccard_bag_similarity.TextReuseTextDocument <- function(a, b) {
  assert_that(all(class(a) == class(b)))
  jaccard_bag_similarity(a$hashes, b$hashes)
}

#' @export
#' @rdname similarity-functions
ratio_of_matches <- function(a, b) UseMethod("ratio_of_matches")

#' @export
ratio_of_matches.default <- function(a, b) {
  assert_that(all(class(a) == class(b)))
  sum(b %in% a) / length(b)
}

#' @export
ratio_of_matches.TextReuseTextDocument <- function(a, b) {
  assert_that(all(class(a) == class(b)))
  ratio_of_matches(a$hashes, b$hashes)
}

#' @export
#' @rdname similarity-functions
count_matches <- function(a, b) UseMethod("count_matches")

#' @export
count_matches.default <- function(a, b) {
  length(matching_tokens(a, b))
}

#' @export
count_matches.TextReuseTextDocument <- function(a, b) {
  assert_that(all(class(a) == class(b)))
  count_matches(a$hashes, b$hashes)
}

#' @export
#' @rdname similarity-functions
matching_tokens <- function(a, b) UseMethod("matching_tokens")

#' @export
matching_tokens.default <- function(a, b) {
  assert_that(all(class(a) == class(b)))
  b[b %in% a]
}

#' @export
matching_tokens.TextReuseTextDocument <- function(a, b) {
  assert_that(all(class(a) == class(b)),
              has_tokens(a),
              has_tokens(b))
  matching_tokens(a$tokens, b$tokens)
}

context("Ratio of matches")

test_that("calculates the value correctly", {
  expect_equal(ratio_of_matches(1:4, 3:5), 2/3)
  expect_equal(ratio_of_matches(letters[1:3], letters[3:4]), 0.5)
  a <- sample(letters, 10, replace = TRUE)
  b <- sample(letters, 10, replace = TRUE)
  res <- ratio_of_matches(a, b)
  expect_true(0 <= res & res <= 1)
})

test_that("counts matches directly", {
  expect_equal(count_matches(1:4, 3:5), 2L)
  expect_equal(count_matches(letters[1:3], letters[3:4]), 1L)
  expect_equal(count_matches(c("a", "b"), c("a", "a", "c")), 2L)
})

test_that("returns matching tokens", {
  expect_equal(matching_tokens(c("a", "b"), c("a", "a", "c")), c("a", "a"))
  expect_equal(matching_tokens(1:4, 3:5), 3:4)
})

test_that("works with TextReuseTextDocument", {
  ny <- system.file("extdata/legal/ny1850-match.txt", package = "textreuse")
  ca <- system.file("extdata/legal/ca1851-match.txt", package = "textreuse")
  ny <- TextReuseTextDocument(ny, meta = list(id = "ny"))
  ca <- TextReuseTextDocument(ca, meta = list(id = "ca"))
  expect_is(jaccard_similarity(ny, ca), "numeric")
})

test_that("counts and returns matching TextReuseTextDocument tokens", {
  a <- TextReuseTextDocument("one two three four",
                             meta = list(id = "a"),
                             tokenizer = tokenize_ngrams,
                             n = 2,
                             keep_tokens = TRUE)
  b <- TextReuseTextDocument("zero one two three",
                             meta = list(id = "b"),
                             tokenizer = tokenize_ngrams,
                             n = 2,
                             keep_tokens = TRUE)
  expect_equal(count_matches(a, b), 2L)
  expect_equal(matching_tokens(a, b), c("one two", "two three"))
})

context("Pairwise comparison")

dir <- system.file("extdata/legal", package = "textreuse")
corpus <- TextReuseCorpus(dir = dir)
cf1 <- pairwise_compare(corpus, jaccard_similarity)
cf2 <- pairwise_compare(corpus, ratio_of_matches, directional = TRUE)

test_that("returns matrix with correct properties", {
  expect_is(cf1, "matrix")
  expect_equal(rownames(cf1), names(corpus))
  expect_is(cf1[1, 3], "numeric")
  expect_is(cf2[3, 1], "numeric")
  expect_equal(cf1[lower.tri(cf1, diag = TRUE)], rep(NA_real_, 6))
  expect_equal(cf2[diag(cf2)], rep(NA_real_, 3))
})

test_that("peforms calculations as expected", {
  expect_equal(cf1[1,2], jaccard_similarity(corpus[[1]], corpus[[2]]))
  expect_equal(cf2[3,2], ratio_of_matches(corpus[[3]], corpus[[2]]))
})

context("Jaccard coefficients")

test_that("calculates the similarity coefficient correctly", {
  expect_equal(jaccard_similarity(1:3, 2:4), 0.5)
  expect_equal(jaccard_similarity(letters[1:3], letters[2:3]), 2/3)
  a <- sample(letters, 10, replace = TRUE)
  b <- sample(letters, 10, replace = TRUE)
  res <- jaccard_similarity(a, b)
  expect_true(0 <= res & res <= 1)
  })

test_that("calculates the dissimilarity coefficient correctly", {
  expect_equal(1 - jaccard_similarity(1:10, 8:11),
               jaccard_dissimilarity(1:10, 8:11))
  expect_equal(1 - jaccard_similarity(letters[1:5], letters[3:10]),
               jaccard_dissimilarity(letters[1:5], letters[3:10]))
})

test_that("works with TextReuseTextDocument", {
  ny <- system.file("extdata/legal/ny1850-match.txt", package = "textreuse")
  ca <- system.file("extdata/legal/ca1851-match.txt", package = "textreuse")
  ny <- TextReuseTextDocument(file = ny)
  ca <- TextReuseTextDocument(file = ca)
  expect_is(jaccard_similarity(ny, ca), "numeric")
})

test_that("calculates bag similarity properly", {
  a <- c("a", "a", "a", "b")
  b <- c("a", "a", "b", "b", "c")
  expect_equal(jaccard_bag_similarity(a, b), 1/3)
})

# Check if the option `mc.cores` has been set. If it has, return `mclapply`
# instead of `lapply`. But in no circumstances use `mclapply` on Windows.
using_parallel <- function() {
  cores_set <- !is.null(getOption("mc.cores"))
  windows <- .Platform$OS.type == "windows"
  cores_set && !windows
}

get_apply_function <- function() {
 if (using_parallel())
   return(parallel::mclapply)
 else
   return(lapply)
}

#' Query a LSH cache for matches to a single document
#'
#' This function retrieves the matches for a single document from an \code{lsh_buckets} object created by \code{\link{lsh}}. See \code{\link{lsh_candidates}} to retrieve all pairs of matches.
#'
#' @param buckets An \code{lsh_buckets} object created by \code{\link{lsh}}.
#' @param id The document ID to find matches for.
#'
#' @return An \code{lsh_candidates} data frame with matches to the document specified.
#'
#' @examples
#' dir <- system.file("extdata/legal", package = "textreuse")
#' minhash <- minhash_generator(200, seed = 235)
#' corpus <- TextReuseCorpus(dir = dir,
#'                           tokenizer = tokenize_ngrams, n = 5,
#'                           minhash_func = minhash)
#' buckets <- lsh(corpus, bands = 50)
#' lsh_query(buckets, "ny1850-match")
#'
#' @seealso \code{\link{lsh}}, \code{\link{lsh_candidates}}
#' @export
lsh_query <- function(buckets, id) {
  assert_that(is_lsh_buckets(buckets),
              is.string(id))

  signatures <- buckets %>%
    dplyr::filter(.data$doc == id) %>%
    `$`("buckets")

  docs <- buckets %>%
    dplyr::filter(.data$buckets %in% signatures) %>%
    `$`("doc")

  res <- tibble::tibble(a = id, b = docs, score = NA_real_) %>%

    dplyr::filter(.data$a != .data$b) %>%
    dplyr::distinct(.data$a, .data$b)

  class(res) <- c("textreuse_candidates", class(res))

  res
}

#' Probability that a candidate pair will be detected with LSH
#'
#' Functions to help choose the correct parameters for the \code{\link{lsh}} and
#' \code{\link{minhash_generator}} functions. Use \code{lsh_threshold} to
#' determine the minimum Jaccard similarity for two documents for them to likely
#' be considered a match. Use \code{lsh_probability} to determine the
#' probability that a pair of documents with a known Jaccard similarity will be
#' detected.
#'
#' @param h The number of minhash signatures.
#' @param b The number of LSH bands.
#' @param s The Jaccard similarity.
#' @details Locality sensitive hashing returns a list of possible matches for
#' similar documents. How likely is it that a pair of documents will be detected
#' as a possible match? If \code{h} is the number of minhash signatures,
#' \code{b} is the number of bands in the LSH function (implying then that the
#' number of rows \code{r = h / b}), and \code{s} is the actual Jaccard
#' similarity of the two documents, then the probability \code{p} that the two
#' documents will be marked as a candidate pair is given by this equation.
#'
#' \deqn{p = 1 - (1 - s^{r})^{b}}
#'
#' According to \href{http://infolab.stanford.edu/~ullman/mmds/book.pdf}{MMDS},
#' that equation approximates an S-curve. This implies that there is a threshold
#' (\code{t}) for \code{s} approximated by this equation.
#'
#' \deqn{t = \frac{1}{b}^{\frac{1}{r}}}
#'
#' @references Jure Leskovec, Anand Rajaraman, and Jeff Ullman,
#'  \emph{Mining of Massive Datasets} (Cambridge University Press, 2011), ch. 3.
#' @examples
#' # Threshold for default values
#' lsh_threshold(h = 200, b = 40)
#'
#' # Probability for varying values of s
#' lsh_probability(h = 200, b = 40, s = .25)
#' lsh_probability(h = 200, b = 40, s = .50)
#' lsh_probability(h = 200, b = 40, s = .75)
#' @export
lsh_probability <- function(h, b, s) {
  assert_that(is.count(h),
              is.count(b),
              check_banding(h, b),
              is.number(s))
  1 - (1 - s ^ (h / b)) ^ b
}

#' @rdname lsh_probability
#' @export
lsh_threshold <- function(h, b) {
  assert_that(is.count(h),
              is.count(b),
              check_banding(h, b))
  (1 / b ) ^ (1 / (h / b))
}

#' @details
#' The best place to begin with this package in the introductory vignette.
#'
#' \code{vignette("textreuse-introduction", package = "textreuse")}
#'
#' After reading that vignette, the "pairwise" and "minhash" vignettes introduce
#' specific paths for working with the package.
#'
#' \code{vignette("textreuse-pairwise", package = "textreuse")}
#'
#' \code{vignette("textreuse-minhash", package = "textreuse")}
#'
#' \code{vignette("textreuse-alignment", package = "textreuse")}
#'
#' Another good place to begin with the package is the documentation for loading
#' documents (\code{\link{TextReuseTextDocument}} and
#' \code{\link{TextReuseCorpus}}), for \link{tokenizers},
#' \link[=similarity-functions]{similarity functions}, and
#' \link[=lsh]{locality-sensitive hashing}.
#'
#' @references The sample data provided in the \code{extdata/ats} directory
#'   contains nineteenth-century American Tract Society publications gathered
#'   from the \href{https://archive.org/}{Internet Archive}.
#'
#'   The sample data provided in the \code{extdata/legal} directory, are taken
#'   from the following nineteenth-century codes of civil procedure from
#'   California and New York.
#'
#'   \emph{Final Report of the Commissioners on Practice and Pleadings}, in 2
#'   \emph{Documents of the Assembly of New York}, 73rd Sess., No. 16, (1850):
#'   243-250, sections 597-613.
#'   \href{http://books.google.com/books?id=9HEbAQAAIAAJ&pg=PA243#v=onepage&q&f=false}{Google
#'    Books}.
#'
#'   \emph{An Act To Regulate Proceedings in Civil Cases}, 1851 \emph{California
#'   Laws} 51, 51-53 sections 4-17; 101, sections 313-316.
#'   \href{http://books.google.com/books?id=4PHEAAAAIAAJ&pg=PA51#v=onepage&q&f=false}{Google
#'    Books}.
#'
#' @useDynLib textreuse, .registration = TRUE
#' @importFrom Rcpp sourceCpp
#' @import RcppProgress
#' @import stringr
#' @import assertthat
#' @importFrom utils getTxtProgressBar setTxtProgressBar txtProgressBar
"_PACKAGE"

if (getRversion() >= "2.15.1") {
  utils::globalVariables(c("doc.x", "doc.y", "up", "dn", "a", "b", ".data", "band", "hash"))
}

#' TextReuseTextDocument
#'
#' This is the constructor function for \code{TextReuseTextDocument} objects.
#' This class is used for comparing documents.
#'
#' @param text A character vector containing the text of the document. This
#'   argument can be skipped if supplying \code{file}.
#' @param file The path to a text file, if \code{text} is not provided.
#' @param meta A list with named elements for the metadata associated with this
#'   document. If a document is created using the \code{text} parameter, then
#'   you must provide an \code{id} field, e.g., \code{meta = list(id =
#'   "my_id")}. If the document is created using \code{file}, then the ID will
#'   be created from the file name.
#' @param tokenizer A function to split the text into tokens. See
#'   \code{\link{tokenizers}}. If value is \code{NULL}, then tokenizing and
#'   hashing will be skipped.
#' @param ... Arguments passed on to the \code{tokenizer}.
#' @param hash_func A function to hash the tokens. See
#'   \code{\link{hash_string}}.
#' @param minhash_func A function to create minhash signatures of the document.
#'   See \code{\link{minhash_generator}}.
#' @param keep_tokens Should the tokens be saved in the document that is
#'   returned or discarded?
#' @param keep_text Should the text be saved in the document that is returned or
#'   discarded?
#' @param skip_short Should short documents be skipped? (See details.)
#' @param encoding Encoding to be used when reading files.
#'
#' @details This constructor function follows a three-step process. It reads in
#'   the text, either from a file or from memory. It then tokenizes that text.
#'   Then it hashes the tokens. Most of the comparison functions in this package
#'   rely only on the hashes to make the comparison. By passing \code{FALSE} to
#'   \code{keep_tokens} and \code{keep_text}, you can avoid saving those
#'   objects, which can result in significant memory savings for large corpora.
#'
#'   If \code{skip_short = TRUE}, this function will return \code{NULL} for very
#'   short or empty documents. A very short document is one where there are too
#'   few words to create at least two n-grams. For example, if five-grams are
#'   desired, then a document must be at least six words long. If no value of
#'   \code{n} is provided, then the function assumes a value of \code{n = 3}. A
#'   warning will be printed with the document ID of a skipped document.
#'
#' @return An object of class \code{TextReuseTextDocument}. This object inherits
#'   from the virtual S3 class \code{\link[NLP]{TextDocument}} in the NLP
#'   package. It contains the following elements: \describe{ \item{content}{The
#'   text of the document.} \item{tokens}{The tokens created from the text.}
#'   \item{hashes}{Hashes created from the tokens.} \item{minhashes}{The minhash
#'   signature of the document.} \item{metadata}{The document metadata,
#'   including the filename (if any) in \code{file}.} }
#'
#' @seealso \link[=TextReuseTextDocument-accessors]{Accessors for TextReuse
#'   objects}.
#'
#' @examples
#' file <- system.file("extdata/legal/ny1850-match.txt", package = "textreuse")
#' doc  <- TextReuseTextDocument(file = file, meta = list(id = "ny1850"))
#' print(doc)
#' meta(doc)
#' head(tokens(doc))
#' head(hashes(doc))
#' \dontrun{
#' content(doc)
#' }
#' @export
TextReuseTextDocument <- function(text, file = NULL, meta = list(),
                                  tokenizer = tokenize_ngrams, ...,
                                  hash_func = hash_string,
                                  minhash_func = NULL,
                                  keep_tokens = FALSE,
                                  keep_text = TRUE,
                                  skip_short = TRUE,
                                  encoding = "unknown") {

  if (!missing(text)) assert_that(has_id(meta))

  if (!is.null(file)) {
    assert_that(missing(text),
                is.readable(file))
    text <- as_string(readLines(file, encoding = encoding))
  }

  assert_that(is.character(text))
  text <- as_string(text)

  # Define document ID early
  document_id <- ifelse(is.null(meta$id), filenames(file), meta$id)

  # Check length of document
  if (skip_short) {
    min_words <- short_document_word_minimum(tokenizer, list(...))
    if (wordcount(text) < min_words) {
      warning("Skipping document with ID '", document_id,
              "' because it has too few words ",
              "to create tokens with the requested tokenizer.",
              call. = FALSE, noBreaks. = TRUE)
      return(NULL)
    }
  }

  # Tokenize and hash
  if (!is.null(tokenizer)) {

    assert_that(is.function(tokenizer))
    tokenizer_name <- as.character(substitute(tokenizer))
    tokens <- tokenizer(text, ...)

    assert_that(is.function(hash_func))
    hash_func_name <- as.character(substitute(hash_func))
    hashes <- hash_func(tokens)

    # Also minhash if requested
    if (!is.null(minhash_func)) {
      assert_that(is.function(minhash_func))
      minhash_func_name <- as.character(substitute(minhash_func))
      minhashes <- minhash_func(tokens)
    } else {
      minhashes <- NULL
      minhash_func_name <- NULL
    }

  } else {
    tokens <- NULL
    hashes <- NULL
    minhashes <- NULL
    tokenizer_name <- NULL
    hash_func_name <- NULL
    minhash_func_name <- NULL

  }

  if (!keep_tokens) tokens <- NULL
  if (!keep_text) text <- NULL

  if (missing(meta)) {
    meta <- list(file = file,
                 id = document_id,
                 tokenizer = tokenizer_name,
                 hash_func = hash_func_name,
                 minhash_func = minhash_func_name)
  }
  assert_that(is.list(meta))
  if (!is.null(file)) {
    meta$file <- file
    meta$id <- document_id
  }
  # Don't overwrite these when called from TextReuseCorpus
  if (is.null(meta$tokenizer) & is.null(meta$hash_func) &
      is.null(meta$minhash_func)) {
    meta$tokenizer <- tokenizer_name
    meta$hash_func <- hash_func_name
    meta$minhash_func <- minhash_func_name
  }

  meta <- sort_meta(meta)

  doc <- list(
    content   = text,
    tokens    = tokens,
    hashes    = hashes,
    minhashes = minhashes,
    meta      = meta
    )

  class(doc) <- c("TextReuseTextDocument", "TextDocument")

  doc

}

short_document_word_minimum <- function(tokenizer, args) {
  n <- args$n
  if (is.null(n)) n <- 3

  if (!is.null(tokenizer) && identical(tokenizer, tokenize_skip_ngrams)) {
    k <- args$k
    if (is.null(k)) k <- 1
    return(n + n * k - k)
  }

  n + 1
}

#' @importFrom NLP meta
#' @export
NLP::meta

#' @importFrom NLP meta<-
#' @export
NLP::`meta<-`

#' @importFrom NLP content
#' @export
NLP::content

#' @importFrom NLP content<-
#' @export
NLP::`content<-`

#' @export
print.TextReuseTextDocument <- function(x, ...) {
  cat("TextReuseTextDocument\n")
  pretty_print_metadata(x)
  cat("content", ":", str_sub(x$content, end = 200))
  invisible(x)
}

#' @export
as.character.TextReuseTextDocument <- function(x, ...) {
  as.character(x$content)
}

#' @export
#' @method content TextReuseTextDocument
content.TextReuseTextDocument <- function(x) {
  x$content
}

#' @export
#' @method content<- TextReuseTextDocument
`content<-.TextReuseTextDocument` <- function(x, value) {
  assert_that(is.character(value))
  x$content <- value
  x
}

#' @export
#' @method meta TextReuseTextDocument
meta.TextReuseTextDocument <- function(x, tag = NULL, ...) {
  if (is.null(tag))
    x$meta
  else
    x$meta[[tag]]
}

#' @export
#' @method meta<- TextReuseTextDocument
`meta<-.TextReuseTextDocument` <- function(x, tag = NULL, ..., value) {
  if (is.null(tag)) {
    assert_that(is.list(value))
    x$meta <- value
  } else {
    x$meta[[tag]] <- value
  }
  x
}

#' Accessors for TextReuse objects
#'
#' Accessor functions to read and write components of
#' \code{\link{TextReuseTextDocument}} and \code{\link{TextReuseCorpus}}
#' objects.
#' @name TextReuseTextDocument-accessors
#' @param x The object to access.
#' @param value The value to assign.
#' @return Either a vector or a named list of vectors.
NULL

#' @export
#' @rdname TextReuseTextDocument-accessors
tokens <- function(x) UseMethod("tokens", x)

#' @export
tokens.TextReuseTextDocument <- function(x) x$tokens

#' @export
tokens.TextReuseCorpus <- function(x) {
  corpus_names <- names(x)
  l <- lapply(x$documents, function(i) tokens(i))
  names(l) <- corpus_names
  l
}

#' @export
#' @rdname TextReuseTextDocument-accessors
`tokens<-` <- function(x, value) UseMethod("tokens<-", x)

#' @export
`tokens<-.TextReuseTextDocument` <- function(x, value) {
  x$tokens <- value
  x
}

#' @export
#' @rdname TextReuseTextDocument-accessors
hashes <- function(x) UseMethod("hashes", x)

#' @export
hashes.TextReuseTextDocument <- function(x) x$hashes

#' @export
hashes.TextReuseCorpus <- function(x) {
  corpus_names <- names(x)
  l <- lapply(x$documents, function(i) hashes(i))
  names(l) <- corpus_names
  l
}

#' @export
#' @rdname TextReuseTextDocument-accessors
`hashes<-` <- function(x, value) UseMethod("hashes<-", x)

#' @export
`hashes<-.TextReuseTextDocument` <- function(x, value) {
  x$hashes <- value
  x
}

#' @export
#' @rdname TextReuseTextDocument-accessors
minhashes <- function(x) UseMethod("minhashes", x)

#' @export
minhashes.TextReuseTextDocument <- function(x) x$minhashes

#' @export
minhashes.TextReuseCorpus <- function(x) {
  corpus_names <- names(x)
  l <- lapply(x$documents, function(i) minhashes(i))
  names(l) <- corpus_names
  l
}

#' @export
#' @rdname TextReuseTextDocument-accessors
`minhashes<-` <- function(x, value) UseMethod("minhashes<-", x)

#' @export
`minhashes<-.TextReuseTextDocument` <- function(x, value) {
  x$minhashes <- value
  x
}

#' @param x An R object to check.
#' @export
#' @rdname TextReuseTextDocument
is.TextReuseTextDocument <- function(x) {
  inherits(x, "TextReuseTextDocument")
}

#' @export
#' @rdname TextReuseTextDocument
has_content <- function(x) {
  assert_that(is.TextReuseTextDocument(x))
  !is.null(x$content)
}

assertthat::on_failure(has_content) <- function(call, env) {
  paste0("Document does not have text in its content field.")
}

#' @export
#' @rdname TextReuseTextDocument
has_tokens <- function(x) {
  assert_that(is.TextReuseTextDocument(x))
  !is.null(x$tokens)
}

assertthat::on_failure(has_tokens) <- function(call, env) {
  "Document does not have tokens."
}

#' @export
#' @rdname TextReuseTextDocument
has_hashes <- function(x) {
  assert_that(is.TextReuseTextDocument(x))
  !is.null(x$hashes)
}

assertthat::on_failure(has_hashes) <- function(call, env) {
  "Document does not have hashes."
}

#' @export
#' @rdname TextReuseTextDocument
has_minhashes <- function(x) {
  assert_that(is.TextReuseTextDocument(x))
  !is.null(x$minhashes)
}

assertthat::on_failure(has_minhashes) <- function(call, env) {
  "Document does not have a minhash signature."
}

has_minhashes_corpus <- function(x) {
  assert_that(is.TextReuseCorpus(x))
  all(vapply(minhashes(x), Negate(is.null), logical(1)))
}

assertthat::on_failure(has_minhashes_corpus) <- function(call, env) {
  "Some documents in the corpus do not have a minhash signature."
}

context("Alignment")

test_that("returns correct results with edits properly marked", {
  a <- "How can we tell if this is a match, he asked?"
  b <- "Surely, this is a good match, she replied."
  res <- align_local(a, b)
  expect_equal(res$a_edits, "this is a #### match")
  expect_equal(res$b_edits, "this is a good match")
  expect_is(res, "list")
  expect_is(res, "textreuse_alignment")
})

test_that("works with TextReuseTextDocuments", {
  file_a <- system.file("extdata/legal/ca1851-match.txt",
                         package = "textreuse")
  file_b <- system.file("extdata/legal/ny1850-match.txt",
                        package = "textreuse")
  doc_a <- TextReuseTextDocument(file = file_a)
  doc_b <- TextReuseTextDocument(file = file_b)
  doc_res <- align_local(doc_a, doc_b)
  expect_is(doc_res, "textreuse_alignment")
  expect_gt(wordcount(doc_res$a_edits), 200)
})

test_that("prints alignment with a trailing blank line", {
  res <- align_local("this is a match", "this is a match")
  output <- capture.output(print(res))
  expect_equal(tail(output, 1), "")
})

test_that("returns empty alignment when there is no match", {
  expect_warning(res <- align_local("abc", "xyz"), NA)
  expect_equal(res$score, 0)
  expect_equal(res$a_edits, "")
  expect_equal(res$b_edits, "")
})

test_that("can preserve punctuation in alignment output", {
  res <- align_local("Hello, world! This is a match.",
                     "hello world this is a match",
                     preserve_punctuation = TRUE)
  expect_equal(res$a_edits, "Hello, world! This is a match.")
  expect_equal(res$b_edits, "hello world this is a match")
})

context("Minhash")

mhash <- minhash_generator()
file <- system.file("extdata/legal/ny1850-match.txt", package = "textreuse")
doc <- TextReuseTextDocument(file = file, hash_func = mhash, keep_tokens = TRUE)

test_that("random integers can be generated", {
  expect_is(random_ints(3), "integer")
  expect_equal(length(random_ints(5)), 5)
})

test_that("minhash generator returns a proper function", {
  expect_is(mhash, "function")
  e <- environment(mhash)
  expect_is(e$r, "integer")
  expect_equal(length(e$r), 200)
})

test_that("minhash works on a TextReuseTextDocument", {
  expect_equal(length(hashes(doc)), 200)
  expect_is(hashes(doc), "integer")
  expect_equal(mhash(tokens(doc)), hashes(doc))
})

test_that("minhash_generator can take seeds and generate same results", {
  mhash_a <- minhash_generator(n = 1000, seed = -5633)
  mhash_b <- minhash_generator(n = 1000, seed = -5633)
  mhash_c <- minhash_generator(n = 1000, seed = 12)
  expect_equal(mhash_a(content(doc)), mhash_b(content(doc)))
  expect_equal(mhash_a(content(doc)) == mhash_c(content(doc)), rep(FALSE, 1000))
})

context("Word counts")

dir <- system.file("extdata/legal", package = "textreuse")
corpus <- TextReuseCorpus(dir = dir)

test_that("counts words correctly for different classes", {
  w <- c("One two three four five six seven; all good children go to heaven.")
  expect_equal(wordcount(w), 13)
  w_doc <- TextReuseTextDocument(w, meta = list(id = "test"))
  expect_equal(wordcount(w), 13)
})

test_that("counts words for a corpus", {
  wc <- wordcount(corpus)
  expect_true(!is.null(names(wc)))
  wc <- unname(wc)
})

context("TextReuseCorpus")

ny <- system.file("extdata/legal/ny1850-match.txt", package = "textreuse")
ca1 <- system.file("extdata/legal/ca1851-nomatch.txt", package = "textreuse")
ca2 <- system.file("extdata/legal/ca1851-match.txt", package = "textreuse")
paths <- c(ca2, ca1, ny)
dir <- system.file("extdata/legal", package = "textreuse")
meta <- list("Corpus name" = "B")
corpus_a <- TextReuseCorpus(paths = paths, keep_tokens = TRUE)
corpus_b <- TextReuseCorpus(dir = dir, keep_tokens = TRUE)

test_that("loads from paths or directory identically", {
  expect_identical(corpus_a, corpus_b)
})

test_that("has metadata", {
  expect_is(meta(corpus_a), "list")
  meta(corpus_b) <- meta
  expect_identical(meta(corpus_b), meta)
})

test_that("has accessor functions", {
  expect_equal(names(corpus_a), filenames(paths))
  names(corpus_b) <- letters[1:3]
  expect_equal(names(corpus_b), letters[1:3])
  expect_equal(length(corpus_a), 3)
})

test_that("can read corpus files with an explicit encoding", {
  dir <- tempfile("encoded-corpus")
  dir.create(dir)
  on.exit(unlink(dir, recursive = TRUE))
  text <- enc2utf8("Café crème is naïve text.")
  writeBin(charToRaw(paste0(text, "\n")), file.path(dir, "encoded.txt"))

  corpus <- TextReuseCorpus(dir = dir, encoding = "UTF-8", tokenizer = NULL)
  expect_equal(as.character(content(corpus[["encoded"]])), text)
  expect_equal(Encoding(as.character(content(corpus[["encoded"]]))), "UTF-8")
})

test_that("has the right classes", {
  expect_is(corpus_a, "TextReuseCorpus")
  expect_is(corpus_a, "Corpus")
})

test_that("has subset methods", {
  expect_identical(corpus_a[[1]],
                   TextReuseTextDocument(file = paths[1], keep_tokens = TRUE))
  # by file path
  expect_identical(corpus_a[[filenames(paths[3])]],
                   TextReuseTextDocument(file = paths[3], keep_tokens = TRUE))
  expect_equal(length(corpus_a[2:3]), 2)
  expect_is(corpus_a[2:3], "TextReuseCorpus")
})

test_that("prints sensibly", {
  expect_output(print(corpus_a), "TextReuseCorpus")
  expect_output(print(corpus_a), "Number of documents: 3")
})

test_that("can be retokenized", {
  expect_equal(tokens(corpus_a[[1]])[1:2],
               tokenize_ngrams(content(corpus_a[[1]]))[1:2])
  corpus_a <- tokenize(corpus_a, tokenize_words)
  expect_equal(tokens(corpus_a[[1]])[1:2],
               tokenize_words(content(corpus_a[[1]]))[1:2])
})

test_that("has methods for tokens and hashes", {
  t <- tokens(corpus_a)
  h <- hashes(corpus_b)
  expect_is(h, "list")
  expect_is(t, "list")
  expect_is(t[[2]], "character")
  expect_is(h[[1]], "integer")
  expect_named(t, names(corpus_a))
  expect_named(h, names(corpus_b))
})

test_that("can create corpus from a character vector with or without IDs", {
  doc_vec <- c("This is document one", "This is document two")
  doc_vec_named <- doc_vec
  names(doc_vec_named) <- c("One", "Two")
  corpus_from_vec <- TextReuseCorpus(text = doc_vec, tokenize = NULL)
  corpus_from_named <- TextReuseCorpus(text = doc_vec_named, tokenize = NULL)
  expect_equal(length(corpus_from_vec), 2)
  expect_equal(length(corpus_from_named), 2)
  expect_equal(as.character(content(corpus_from_vec[[2]])),
               "This is document two")
  expect_equal(as.character(content(corpus_from_named[[1]])),
               "This is document one")
  expect_equal(meta(corpus_from_vec[[1]], "id"), "doc-1")
  expect_equal(meta(corpus_from_named[[2]], "id"), "Two")
  expect_equal(corpus_from_vec[["doc-2"]], corpus_from_vec[[2]])
  expect_equal(corpus_from_named[["One"]], corpus_from_named[[1]])
  expect_error(TextReuseCorpus(text = "Document", dir = "/tmp"))
})

test_that("skips documents that are too short", {
  texts <- c("short" = "Too short", "long" = "Just long enough yo")
  expect_warning(short_docs <- TextReuseCorpus(text = texts, skip_short = TRUE),
                 "Skipping document with ID")
  expect_lt(length(short_docs), length(texts))
  expect_equal(skipped(short_docs), "short")
})

test_that("skipped IDs are always available", {
  texts <- c("short-a" = "Too short",
             "long" = "Just long enough yo",
             "short-b" = "Also short")
  expect_warning(short_docs <- TextReuseCorpus(text = texts, skip_short = TRUE),
                 "Skipping document with ID")
  expect_equal(skipped(short_docs), c("short-a", "short-b"))

  all_docs <- TextReuseCorpus(text = texts, tokenizer = NULL,
                              skip_short = FALSE)
  expect_equal(skipped(all_docs), character())
})

test_that("skips documents that are too short for skip n-grams", {
  texts <- c("short" = "one two three four five",
             "long" = "one two three four five six seven")
  expect_warning(short_docs <- TextReuseCorpus(text = texts,
                                               tokenizer = tokenize_skip_ngrams,
                                               n = 3, k = 2,
                                               skip_short = TRUE),
                 "Skipping document with ID")
  expect_equal(length(short_docs), 1)
  expect_equal(names(short_docs), "long")
})

test_that("gives warning when skipping short documents from files", {
  skip_on_cran()
  too_short <- tempfile("tooshort", fileext = ".txt")
  too_short2 <- tempfile("tooshort2", fileext = ".txt")
  writeLines("Two words", too_short)
  writeLines("Two words", too_short2)
  expect_warning(short <- TextReuseCorpus(paths = c(too_short, too_short2), n = 5),
                 "Skipping document with ID 'tooshort")
  expect_equal(length(short), 0)
  file.remove(too_short)
})

context("TextReuseTextDocument")

doc <- TextReuseTextDocument(file = "newman.txt", keep_tokens = TRUE)
test_meta <- list(id = "test")

test_that("inherits from the correct classes", {
  expect_is(doc, c("TextReuseTextDocument", "TextDocument"))
})

test_that("has the correct structure", {
  expect_named(doc, c("content", "tokens", "hashes", "minhashes", "meta"))
})

test_that("can set the metadata", {
  expect_named(meta(doc), c("file", "hash_func", "id", "minhash_func",
                            "tokenizer"))
  doc2 <- TextReuseTextDocument(file = "newman.txt",
                                meta = list(author = "Newman, John Henry"))
  expect_named(meta(doc2), c("author", "file", "hash_func", "id", "tokenizer"))
  expect_equal(meta(doc2, "author"), "Newman, John Henry")
})

test_that("provides the necessary methods", {
  expect_is(as.character(doc), c("String", "character"))
  expect_is(content(doc), "String", "character")
  expect_output(print(doc), "And now that I am about to trace")
  expect_is(meta(doc), c("list"))

  doc3 <- TextReuseTextDocument(file = "newman.txt")
  meta(doc3)    <- list("author" = "Newman, John Henry")
  expect_equal(meta(doc3), list("author" = "Newman, John Henry"))
  content(doc3) <- "Replacing content"
  expect_equal(content(doc3), "Replacing content")
  meta(doc3, "author") <- "Cardinal Newman"
  expect_equal(meta(doc3, "author"), "Cardinal Newman")
})

test_that("has correct tokens", {
  expect_equal(head(doc$tokens, 3),
               c("and now that", "now that i", "that i am"))
})

test_that("can be created from a character vector not just a file", {
  text <- "This is the text of the document."
  doc <- TextReuseTextDocument(text, meta = test_meta)
  expect_equal(text, as.character(doc))
})

test_that("can read files with an explicit encoding", {
  text <- enc2utf8("Café crème is naïve text.")
  file <- tempfile(fileext = ".txt")
  writeBin(charToRaw(paste0(text, "\n")), file)
  on.exit(file.remove(file))

  doc <- TextReuseTextDocument(file = file, encoding = "UTF-8",
                               tokenizer = NULL)
  expect_equal(as.character(doc), text)
  expect_equal(Encoding(as.character(doc)), "UTF-8")
})

test_that("can be retokenized", {
  text <- "This is the text. But also this."
  a <- TextReuseTextDocument(text, meta = test_meta, tokenizer = tokenize_words,
                             keep_tokens = TRUE)
  b <- tokenize(a, tokenizer = tokenize_sentences, keep_tokens = TRUE)
  expect_false(identical(tokens(a), tokens(b)))
  expect_false(identical(hashes(a), hashes(b)))
  expect_equal(tokens(a), c("this", "is", "the", "text", "but", "also", "this"))
  expect_equal(tokens(b), c("this is the text", "but also this"))
})

test_that("can rehash if it has tokens", {
  tokenless <- TextReuseTextDocument(file = "newman.txt")
  expect_true(has_tokens(doc))
  expect_true(!has_tokens(tokenless))
  expect_is(hashes(rehash(doc, hash_string)), "integer")
  expect_error(rehash(tokenless), "tokens")
})

test_that("can be created with no tokens", {
  doc <- TextReuseTextDocument(file = "newman.txt", tokenizer = NULL)
  expect_false(has_tokens(doc))
  expect_false(has_hashes(doc))
})

test_that("skips documents that are too short", {
  expect_warning(short_doc <-
                   TextReuseTextDocument(text = "Too short",
                                         meta = list(id = "short"),
                                         skip_short = TRUE),
                 "Skipping document with ID")
  expect_null(short_doc)
})

test_that("skips documents that are too short for skip n-grams", {
  text <- "one two three four five"
  expect_warning(short_doc <-
                   TextReuseTextDocument(text = text,
                                         meta = list(id = "short-skip"),
                                         tokenizer = tokenize_skip_ngrams,
                                         n = 3, k = 2,
                                         skip_short = TRUE),
                 "Skipping document with ID")
  expect_null(short_doc)
})

test_that("gives warning when skipping short documents from file", {
  skip_on_cran()
  too_short <- tempfile("tooshort", fileext = ".txt")
  writeLines("Two words", too_short)
  expect_warning(short_doc <- TextReuseTextDocument(file = too_short, n = 5),
                 "Skipping document with ID 'tooshort")
  expect_null(short_doc)
  file.remove(too_short)
})

# Take results of readLines and turn it into a character vector of length 1
as_string <- function(x) {
  x %>%
    str_c(collapse = "\n") %>%
    NLP::as.String()
}

# Pretty print the metadata for a document
pretty_print_metadata <- function(doc) {
  lapply(names(doc$meta), function(x) cat(x, ":", doc$meta[[x]], "\n"))
}

# Check whether the number of minhashes is evenly divisible by number of bands
check_banding <- function(h, b) {
  h %% b == 0
}

assertthat::on_failure(check_banding) <- function(call, env) {
  "The number of hashes must be evenly divisible by the number of bands."
}

# Sequences for subsetting by bands in minhash
band_seq <- function(l, b) {
  assert_that(check_banding(l, b))
  r <- l / b
  starts <- seq.int(from = 1, to = l, by = r)
  lapply(starts, function(n) seq.int(n, n + r - 1, 1))
}

# Test that meta exists and that it has an ID value
has_id <- function(meta) {
  !is.null(meta$id)
}

assertthat::on_failure(has_id) <- function(call, env) {
  paste("When creating a document from a string instead of a file, the `id`",
        "field in the metadata list must be specified.")
}

# People might row_bind() two of these data frames, so we can't rely just on
# the class.
is_lsh_buckets <- function(x) {
  identical(names(x), c("doc", "buckets")) & inherits(x, "data.frame")
}

assertthat::on_failure(is_lsh_buckets) <- function(call, env) {
  "Object is not a data frame of LSH buckets."
}

# People might run a candidates data frame through dplyr so that it loses its
# class.
is_candidates_df <- function(x) {
  class_check <- inherits(x, "textreuse_candidates")
  col_check <- all(c("a", "b", "score") %in% names(x)) & inherits(x, "data.frame")
  class_check | col_check
}

assertthat::on_failure(is_candidates_df) <- function(call, env) {
  "Object is not a candidates data frame."
}

is_integer_like <- function(x) {
  is.integer(x) | (is.scalar(x) & (x == as.integer(x)))
}

assertthat::on_failure(is_integer_like) <- function(call, env) {
 paste0(deparse(call$x), " is not a whole number.")
}

sort_meta <- function(meta) {
  meta[order(names(meta))]
}

sort_df_by_rows <- function(df) {
  assert_that(all(c("a", "b") %in% colnames(df)),
              is.data.frame(df))
  for (i in seq_len(nrow(df))) {
    ordered <- sort(c(df[i, "a"], df[i, "b"]))
    df[i, "a"] <- ordered[1]
    df[i, "b"] <- ordered[2]
  }
  df
}

sort_df_by_columns <- function(df) {
  assert_that(all(c("a", "b") %in% colnames(df)),
              is.data.frame(df))
  df <- df[with(df, order(a, b)), ]
  # rownames(df) <- NULL
  df
}

# Given a word, create a string with the same number of marker characters
mark_chars <- function(word, char) {
  str_c(rep(char, str_length(word)), collapse = "")
}

context("Utils")

test_that("as_string returns the correct type", {
  s <- as_string(c("First", "Second"))
  expect_is(s, "String")
  expect_equal(as.character(s), "First\nSecond")
})

test_that("data frames can be sorted by rows", {
  df1 <- data.frame(a = letters[1:3], b = letters[24:26], stringsAsFactors = FALSE)
  df2 <- data.frame(a = letters[24:26], b = letters[1:3], stringsAsFactors = FALSE)
  expect_identical(sort_df_by_rows(df2), df1)
})

test_that("data frames can be sorted by columns", {
  df1 <- data.frame(a = letters[c(2, 3, 1)], b = letters[c(25, 26, 24)],
                    stringsAsFactors = FALSE)
  df2 <- data.frame(a = letters[1:3], b = letters[24:26],
                    stringsAsFactors = FALSE)
  df1 <- textreuse:::sort_df_by_columns(df1)
  rownames(df1) <- NULL
  rownames(df2) <- NULL
  expect_identical(df1, df2)
})

test_that("edit marks can be created correctly", {
  expect_equal(mark_chars("word", "#"), "####")
})

#'Locality sensitive hashing for minhash
#'
#'Locality sensitive hashing (LSH) discovers potential matches among a corpus of
#'documents quickly, so that only likely pairs can be compared.
#'
#'@details Locality sensitive hashing is a technique for detecting document
#'  similarity that does not require pairwise comparisons. When comparing pairs
#'  of documents, the number of pairs grows rapidly, so that only the smallest
#'  corpora can be compared pairwise in a reasonable amount of computation time.
#'  Locality sensitive hashing, on the other hand, takes a document which has
#'  been tokenized and hashed using a minhash algorithm. (See
#'  \code{\link{minhash_generator}}.) Each set of minhash signatures is then
#'  broken into bands comprised of a certain number of rows. (For example, 200
#'  minhash signatures might be broken down into 20 bands each containing 10
#'  rows.) Each band is then hashed to a bucket. Documents with identical rows
#'  in a band will be hashed to the same bucket. The likelihood that a document
#'  will be marked as a potential duplicate is proportional to the number of
#'  bands and inversely proportional to the number of rows in each band.
#'
#'  This function returns a data frame with the additional class
#'  \code{lsh_buckets}. The LSH technique only requires that the signatures for
#'  each document be calculated once. So it is possible, as long as one uses the
#'  same minhash function and the same number of bands, to combine the outputs
#'  from this function at different times. The output can thus be treated as a
#'  kind of cache of LSH signatures.
#'
#'  To extract pairs of documents from the output of this function, see
#'  \code{\link{lsh_candidates}}.
#'
#'@param x A \code{\link{TextReuseCorpus}} or
#'  \code{\link{TextReuseTextDocument}}.
#'@param bands The number of bands to use for locality sensitive hashing. The
#'  number of hashes in the documents in the corpus must be evenly divisible by
#'  the number of bands. See \code{\link{lsh_threshold}} and
#'  \code{\link{lsh_probability}} for guidance in selecting the number of bands
#'  and hashes.
#'@param progress Display a progress bar while comparing documents.
#'
#'@return A data frame (with the additional class \code{lsh_buckets}),
#'  containing a column with the document IDs and a column with their LSH
#'  signatures, or buckets.
#'
#'@references Jure Leskovec, Anand Rajaraman, and Jeff Ullman,
#'  \emph{Mining of Massive Datasets} (Cambridge University Press, 2011), ch.
#'  3. See also Matthew Casperson,
#'  "\href{http://matthewcasperson.blogspot.com/2013/11/minhash-for-dummies.html}{Minhash
#'   for Dummies}" (November 14, 2013).
#'
#'@seealso \code{\link{minhash_generator}}, \code{\link{lsh_add}},
#'  \code{\link{lsh_candidates}}, \code{\link{lsh_query}},
#'  \code{\link{lsh_probability}},
#'  \code{\link{lsh_threshold}}
#'
#' @examples
#' dir <- system.file("extdata/legal", package = "textreuse")
#' minhash <- minhash_generator(200, seed = 235)
#' corpus <- TextReuseCorpus(dir = dir,
#'                           tokenizer = tokenize_ngrams, n = 5,
#'                           minhash_func = minhash)
#' buckets <- lsh(corpus, bands = 50)
#' buckets
#'@export
lsh <- function(x, bands, progress = interactive()) {
  UseMethod("lsh", x)
}

#' Add documents to a LSH cache
#'
#' This function adds buckets for one or more new documents to an existing
#' \code{lsh_buckets} object. Use the same \code{bands} value and minhash
#' function that were used to create the original buckets.
#'
#' @param buckets An \code{lsh_buckets} object created by \code{\link{lsh}}.
#' @param x A \code{\link{TextReuseCorpus}} or
#'   \code{\link{TextReuseTextDocument}} with minhashes.
#' @inheritParams lsh
#' @return An updated \code{lsh_buckets} object.
#' @seealso \code{\link{lsh}}, \code{\link{lsh_query}},
#'   \code{\link{lsh_candidates}}
#' @export
lsh_add <- function(buckets, x, bands, progress = interactive()) {
  assert_that(is_lsh_buckets(buckets))

  new_buckets <- lsh(x, bands = bands, progress = progress)
  new_doc_ids <- unique(new_buckets$doc)

  buckets <- buckets %>%
    dplyr::filter(!.data$doc %in% new_doc_ids) %>%
    dplyr::bind_rows(new_buckets) %>%
    dplyr::arrange(.data$doc)

  class(buckets) <- c("lsh_buckets", setdiff(class(buckets), "lsh_buckets"))

  buckets
}

#' @export
lsh.TextReuseCorpus <- function(x, bands, progress = interactive()) {

  assert_that(is.count(bands),
              has_minhashes_corpus(x))

  h <- length(minhashes(x[[1]])) # number of hashes
  d <- length(x) # number of documents
  r <- h / bands # number of rows

  assert_that(check_banding(h, bands))

  # To assign rows in data frame to bands
  b_assign <-  tibble::tibble(band =
      rep(vapply(1:bands, function(i) rep(i, r), integer(r)), d)
    )

  all_minhashes <- minhashes(x)
  col_names <- names(all_minhashes)

  buckets <- all_minhashes %>%
    tibble::as_tibble() %>%
    tidyr::gather("doc", "hash", col_names) %>%
    dplyr::mutate(doc = as.character(.data$doc)) %>%
    dplyr::bind_cols(b_assign) %>%
    dplyr::group_by(.data$doc, .data$band)

    rm(b_assign)

    if (progress) {
      message("Calculating LSH buckets")
      pb <- txtProgressBar(min = 0, max = d * bands, style = 3)
    }

  # include the band in the signature hash to avoid false matches
  buckets <- buckets %>%
    dplyr::summarize(buckets = digest_progress(list(hash, unique(band)),
                                                 pb, progress))

  if (progress) close(pb)

  buckets <- buckets %>%
    dplyr::select(-.data$band) %>%
    dplyr::ungroup()

  class(buckets) <- c("lsh_buckets", class(buckets))

  buckets

}

# A wrapper around digest to be able to use the progress bar
digest_progress <- function(x, pb, progress) {
  bucket <- digest::digest(x)
  if (progress) setTxtProgressBar(pb, getTxtProgressBar(pb) + 1)
  bucket
}

#' @export
lsh.TextReuseTextDocument <- function(x, bands, progress) {

  assert_that(is.count(bands),
              has_minhashes(x))

  all_minhashes <- minhashes(x)
  h <- length(all_minhashes) # number of hashes
  r <- h / bands # number of rows

  assert_that(check_banding(h, bands))

  # To assign rows in data frame to bands
  b_assign <-  tibble::tibble(band =
      rep(vapply(1:bands, function(i) rep(i, r), integer(r)), 1)
    )


  buckets <- tibble::tibble(doc = x$meta$id, hash = all_minhashes) %>%
    dplyr::bind_cols(b_assign) %>%
    dplyr::group_by(.data$doc, .data$band) %>%
    dplyr::summarize(buckets = digest::digest(list(hash, unique(band)))) %>%
    dplyr::select(-.data$band) %>%
    dplyr::ungroup()

  class(buckets) <- c("lsh_buckets", class(buckets))

  buckets

}

context("LSH")

dir <- system.file("extdata/legal", package = "textreuse")
minhash <- minhash_generator(200, seed = 9228)
corpus <- TextReuseCorpus(dir = dir,
                          tokenizer = tokenize_ngrams, n = 5,
                          keep_tokens = TRUE,
                          minhash_func = minhash)
buckets <- lsh(corpus, bands = 50)
candidates <- lsh_candidates(buckets)
scores <- lsh_compare(candidates, corpus, jaccard_similarity)

test_that("returns a data frame", {
  expect_is(buckets, "tbl_df")
})

test_that("returns error if improper number of bands are chosen", {
  expect_error(lsh(corpus, bands = 33), "The number of hashes")
})

test_that("creates buckets without warnings", {
  expect_warning(lsh(corpus, bands = 50), NA)
})

test_that("returns pairs of candidates without duplicates", {
  expect_is(candidates, "data.frame")
  expect_named(candidates, c("a", "b", "score"))
  expect_equal(candidates[[1, 1]], "ca1851-match")
  expect_equal(candidates[[1, 2]], "ny1850-match")
  expect_equal(candidates[[1, 3]], NA_real_)
  expect_equal(nrow(candidates), 1)
})

test_that("additional documents can be added", {
  corpus2 <- rehash(corpus, minhash, type = "minhashes")
  buckets1and2 <- lsh(corpus2[1:2], bands = 50)
  buckets3 <- lsh(corpus2[[3]], bands = 50)
  buckets_combined <- dplyr::bind_rows(buckets1and2, buckets3)
  expect_equal(buckets_combined$buckets, buckets$buckets)
  expect_equal(buckets_combined$doc, buckets$doc)
})

test_that("lsh buckets can be extended with new documents", {
  corpus2 <- rehash(corpus, minhash, type = "minhashes")
  buckets1and2 <- lsh(corpus2[1:2], bands = 50)
  buckets_added <- lsh_add(buckets1and2, corpus2[[3]], bands = 50)
  expect_is(buckets_added, "lsh_buckets")
  expect_equal(buckets_added$buckets, buckets$buckets)
  expect_equal(buckets_added$doc, buckets$doc)
})

test_that("lsh buckets are replaced when adding an existing document", {
  corpus2 <- rehash(corpus, minhash, type = "minhashes")
  buckets_added <- lsh_add(buckets[1:100, ], corpus2[[1]], bands = 50)
  expect_equal(sum(buckets_added$doc == names(corpus2)[1]), 50)
})

test_that("candidates can be scored", {
  correct <- jaccard_similarity(corpus[["ca1851-match"]],
                                corpus[["ny1850-match"]])
  expect_equal(scores[[1,3]], correct)
})

test_that("pre-scored candidates are not recalculated", {
  candidates_scored <- candidates
  candidates_scored$score[1] <- 1
  rescored <- lsh_compare(candidates_scored, corpus, function(a, b) 0)
  expect_equal(rescored$score[1], 1)
})

test_that("scores can be converted to a matrix", {
  m <- as.matrix(scores)
  expect_is(m, "matrix")
  expect_equal(colnames(m), names(corpus))
})

test_that("can be queried for a single document", {
  match <- lsh_query(buckets, "ca1851-match")
  expect_equal(match$b, "ny1850-match")
  expect_is(match, "textreuse_candidates")
})

#' TextReuseCorpus
#'
#' This is the constructor function for a \code{TextReuseCorpus}, modeled on the
#' virtual S3 class \code{Corpus} from the \code{tm} package. The
#' object is a \code{TextReuseCorpus}, which is basically a list containing
#' objects of class \code{\link{TextReuseTextDocument}}. Arguments are passed
#' along to that constructor function. To create the corpus, you can pass either
#' a character vector of paths to text files using the \code{paths =} parameter,
#' a directory containing text files (with any extension) using the \code{dir =}
#' parameter, or a character vector of documents using the \code{text = }
#' parameter, where each element in the characer vector is a document. If the
#' character vector passed to \code{text = } has names, then those names will be
#' used as the document IDs. Otherwise, IDs will be assigned to the documents.
#' Only one of the \code{paths}, \code{dir}, or \code{text} parameters should be
#' specified.
#'
#' @details If \code{skip_short = TRUE}, this function will skip very short or
#'   empty documents. A very short document is one where there are too few words
#'   to create at least two n-grams. For example, if five-grams are desired,
#'   then a document must be at least six words long. If no value of \code{n} is
#'   provided, then the function assumes a value of \code{n = 3}. A warning will
#'   be printed with the document ID of each skipped document. Use
#'   \code{skipped()} to get the IDs of skipped documents.
#'
#'   This function will use multiple cores on non-Windows machines if the
#'   \code{"mc.cores"} option is set. For example, to use four cores:
#'   \code{options("mc.cores" = 4L)}.
#'
#' @param paths A character vector of paths to files to be opened.
#' @param dir The path to a directory of text files.
#' @param text A character vector (possibly named) of documents.
#' @param meta A list with named elements for the metadata associated with this
#'   corpus.
#' @param progress Display a progress bar while loading files.
#' @param tokenizer A function to split the text into tokens. See
#'   \code{\link{tokenizers}}. If value is \code{NULL}, then tokenizing and
#'   hashing will be skipped.
#' @param ... Arguments passed on to the \code{tokenizer}.
#' @param hash_func A function to hash the tokens. See
#'   \code{\link{hash_string}}.
#' @param minhash_func A function to create minhash signatures of the document.
#'   See \code{\link{minhash_generator}}.
#' @param keep_tokens Should the tokens be saved in the documents that are
#'   returned or discarded?
#' @param keep_text Should the text be saved in the documents that are returned
#'   or discarded?
#' @param skip_short Should short documents be skipped? (See details.)
#' @param encoding Encoding to be used when reading files.
#'
#' @seealso \link[=TextReuseTextDocument-accessors]{Accessors for TextReuse
#'   objects}.
#'
#' @examples
#' dir <- system.file("extdata/legal", package = "textreuse")
#' corpus <- TextReuseCorpus(dir = dir, meta = list("description" = "Field Codes"))
#' # Subset by position or file name
#' corpus[[1]]
#' names(corpus)
#' corpus[["ca1851-match"]]
#'
#' @export
TextReuseCorpus <- function(paths, dir = NULL, text = NULL, meta = list(),
                            progress = interactive(),
                            tokenizer = tokenize_ngrams, ...,
                            hash_func = hash_string,
                            minhash_func = NULL,
                            keep_tokens = FALSE,
                            keep_text = TRUE,
                            skip_short = TRUE,
                            encoding = "unknown") {

  if (!is.null(tokenizer)) {
    assert_that(is.function(tokenizer),
                is.function(hash_func))
    tokenizer_name <- as.character(substitute(tokenizer))
    hash_func_name <- as.character(substitute(hash_func))
    if (!is.null(minhash_func)) {
      minhash_func_name <- as.character(substitute(minhash_func))
    } else {
      minhash_func_name <- NULL
    }
    loading_msg <- "Loading, tokenizing, and hashing "
  } else {
    tokenizer_name <- NULL
    hash_func_name <- NULL
    minhash_func_name <- NULL
    loading_msg <- "Loading "
  }

  apply_func <- get_apply_function()

  # If we get a character vector of documents, use that; otherwise load
  # the files from disk.
  if (!missing(text)) {

    assert_that(missing(paths),
                is.null(dir),
                is.character(text))

    if (progress) {
      len <- length(text)
      message(loading_msg, prettyNum(len, big.mark = ","), " documents.")
      if (using_parallel())
        progress <- FALSE
      else
        pb <- txtProgressBar(min = 0, max = len, style = 3)
    }

    if (is.null(names(text)))
      names(text) <- str_c("doc-", 1:length(text))

    docs <- apply_func(seq_along(text), function(i) {
      d <- TextReuseTextDocument(text = text[i],
                                 tokenizer = tokenizer, ...,
                                 hash_func = hash_func,
                                 minhash_func = minhash_func,
                                 keep_tokens = keep_tokens,
                                 keep_text = keep_text,
                                 skip_short = skip_short,
                                 encoding = encoding,
                                 meta = list(id = names(text)[i],
                                             tokenizer = tokenizer_name,
                                             hash_func = hash_func_name,
                                             minhash_func = minhash_func_name))
      if (progress) setTxtProgressBar(pb, i)
      d
    })

    if (progress) close(pb)

    names(docs) <- names(text)

  } else {

    if (missing(paths) & !is.null(dir)) {
      assert_that(is.dir(dir))
      paths <- Sys.glob(str_c(dir, "/*"))
    }

    vapply(paths, is.readable, logical(1), USE.NAMES = FALSE)

    if (progress) {
      len <- length(paths)
      message(loading_msg, prettyNum(len, big.mark = ","), " documents.")
      if (using_parallel())
        progress <- FALSE
      else
        pb <- txtProgressBar(min = 0, max = len, style = 3)
    }
    docs <- apply_func(seq_along(paths), function(i) {
      d <- TextReuseTextDocument(file = paths[i], tokenizer = tokenizer, ...,
                                 hash_func = hash_func,
                                 minhash_func = minhash_func,
                                 keep_tokens = keep_tokens,
                                 keep_text = keep_text,
                                 skip_short = skip_short,
                                 encoding = encoding,
                                 meta = list(tokenizer = tokenizer_name,
                                             hash_func = hash_func_name,
                                             minhash_func = minhash_func_name))
      if (progress) setTxtProgressBar(pb, i)
      d
    })

    if (progress) close(pb)

    names(docs) <- filenames(paths)
  }

  skipped <- character()

  # Filter documents that were skipped because they were too short
  if (skip_short) {
    skipped_docs <- vapply(docs, is.null, logical(1))
    skipped <- names(docs)[skipped_docs]
    docs <- docs[!skipped_docs]
    if (length(skipped) > 0)
      warning("Skipped ", length(skipped), " documents that were too short. ",
              "Use `skipped()` to get their IDs.")
  }

  assert_that(is.list(meta))
  meta$tokenizer <- tokenizer_name
  meta$hash_func <- hash_func_name
  meta$minhash_func <- minhash_func_name

  if (!is.null(names(meta))) meta <- sort_meta(meta)

  corpus <- list(documents = docs, meta = meta)
  class(corpus) <- c("TextReuseCorpus", "Corpus")
  attr(corpus, "skipped") <- skipped

  corpus

}

#' @export
meta.TextReuseCorpus <- function(x, tag = NULL, ...) {
  if (is.null(tag))
    x$meta
  else
    x$meta[[tag]]
}

#' @export
`meta<-.TextReuseCorpus` <- function(x, tag = NULL, ..., value) {
  if (is.null(tag)) {
    assert_that(is.list(value))
    x$meta <- value
  } else {
    x$meta[[tag]] <- value
  }
  x
}

#' @export
print.TextReuseCorpus <- function(x, ...) {
  cat("TextReuseCorpus\n")
  cat("Number of documents:", length(x), "\n")
  pretty_print_metadata(x)
}

#' @export
length.TextReuseCorpus <- function(x) {
  length(x$documents)
}

#' @export
`[.TextReuseCorpus` <- function(x, i) {
  x$documents <- x$documents[i]
  x
}

#' @export
`[[.TextReuseCorpus` <- function(x, i) {
  x$documents[[i]]
}

#' @export
names.TextReuseCorpus <- function(x) {
  names(x$documents)
}

#' @export
`names<-.TextReuseCorpus` <- function(x, value) {
  names(x$documents) <- value
  x
}

#' @param x An R object to check.
#' @export
#' @rdname TextReuseCorpus
is.TextReuseCorpus <- function(x) {
  inherits(x, "TextReuseCorpus")
}

#' @export
#' @rdname TextReuseCorpus
skipped <- function(x) {
  assert_that(is.TextReuseCorpus(x))
  attr(x, "skipped", exact = TRUE)
}

#' Generate a minhash function
#'
#' A minhash value is calculated by hashing the strings in a character vector to
#' integers and then selecting the minimum value. Repeated minhash values are
#' generated by using different hash functions: these different hash functions
#' are created by using performing a bitwise \code{XOR} operation
#' (\code{\link{bitwXor}}) with a vector of random integers. Since it is vital
#' that the same random integers be used for each document, this function
#' generates another function which will always use the same integers. The
#' returned function is intended to be passed to the \code{hash_func} parameter
#' of \code{\link{TextReuseTextDocument}}.
#'
#' @param n The number of minhashes that the returned function should generate.
#' @param seed An option parameter to set the seed used in generating the random
#'   numbers to ensure that the same minhash function is used on repeated
#'   applications.
#' @return A function which will take a character vector and return \code{n}
#'   minhashes.
#' @references Jure Leskovec, Anand Rajaraman, and Jeff Ullman,
#'   \emph{Mining of Massive Datasets} (Cambridge University Press, 2011), ch.
#'   3. See also Matthew Casperson,
#'   "\href{http://matthewcasperson.blogspot.com/2013/11/minhash-for-dummies.html}{Minhash
#'    for Dummies}" (November 14, 2013).
#' @seealso \code{\link{lsh}}
#' @examples
#' set.seed(253)
#' minhash <- minhash_generator(10)
#'
#' # Example with a TextReuseTextDocument
#' file <- system.file("extdata/legal/ny1850-match.txt", package = "textreuse")
#' doc <- TextReuseTextDocument(file = file, hash_func = minhash,
#'                              keep_tokens = TRUE)
#' hashes(doc)
#'
#' # Example with a character vector
#' is.character(tokens(doc))
#' minhash(tokens(doc))
#' @export
minhash_generator <- function(n = 200, seed = NULL) {
  assert_that(is.count(n))
  if (!is.null(seed)) set.seed(seed)
  r <- random_ints(n)
  f <- function(x) {
    assert_that(is.character(x))
    h <- hash_string(x)
    vapply(r, function(i) { min(bitwXor(h, i)) },
           integer(1), USE.NAMES = FALSE)
  }
  f
}

# Generate random integers for minhashing
#
# It is crucial that you use the same random integers for every document in the
# corpus. The random integers generated by this function are intended to be
# passed to \code{\link{minhash}}.
# @param n The number of random integers to generate.
# @return A vector of integers
# @seealso \code{\link{minhash}}
# @examples
# random_ints(3)
random_ints <- function(n) {
  as.integer(stats::runif(n, -2147483648, 2147483647))
}

#' Candidate pairs from pairwise comparisons
#'
#' Converts a comparison matrix generated by \code{\link{pairwise_compare}} into a
#' data frame of candidates for matches.
#'
#' @param m A matrix from \code{\link{pairwise_compare}}.
#' @param directional Should be set to the same value as in
#'   \code{\link{pairwise_compare}}.
#' @return A data frame containing all the non-\code{NA} values from \code{m}.
#'   Columns \code{a} and \code{b} are the IDs from the original corpus as
#'   passed to the comparison function. Column \code{score} is the score
#'   returned by the comparison function.
#' @examples
#' dir <- system.file("extdata/legal", package = "textreuse")
#' corpus <- TextReuseCorpus(dir = dir)
#'
#' m1 <- pairwise_compare(corpus, ratio_of_matches, directional = TRUE)
#' pairwise_candidates(m1, directional = TRUE)
#'
#' m2 <- pairwise_compare(corpus, jaccard_similarity)
#' pairwise_candidates(m2)
#' @export
pairwise_candidates <- function(m, directional = FALSE) {
  assert_that(is.matrix(m))
  matches <- which(!is.na(m))
  indexes <- arrayInd(matches, dim(m))
  score <- m[matches]
  a <- rownames(m)[indexes[ , 1]]
  b <- colnames(m)[indexes[ , 2]]
  df <- data.frame(a = a, b = b, score = score, stringsAsFactors = FALSE)
  if (!directional) df <- sort_df_by_rows(df)
  df <- sort_df_by_columns(df)
  class(df) <- c("textreuse_candidates", "tbl_df", "tbl", "data.frame")
  df
}

#' Candidate pairs from LSH comparisons
#'
#' Given a data frame of LSH buckets returned from \code{\link{lsh}}, this
#' function returns the potential candidates.
#'
#' @param buckets A data frame returned from \code{\link{lsh}}.
#'
#' @return A data frame of candidate pairs.
#'
#' @examples
#' dir <- system.file("extdata/legal", package = "textreuse")
#' minhash <- minhash_generator(200, seed = 234)
#' corpus <- TextReuseCorpus(dir = dir,
#'                           tokenizer = tokenize_ngrams, n = 5,
#'                           minhash_func = minhash)
#' buckets <- lsh(corpus, bands = 50)
#' lsh_candidates(buckets)
#'
#' @export
lsh_candidates <- function(buckets) {
  assert_that(is_lsh_buckets(buckets))

  candidates <- buckets %>%
    dplyr::left_join(buckets, by = "buckets") %>%
    dplyr::filter(.data$doc.x != .data$doc.y) %>%
    dplyr::distinct(doc.x, doc.y) %>%
    dplyr::arrange(.data$doc.x, .data$doc.y) %>%
    dplyr::mutate(dn = pmin(.data$doc.x, .data$doc.y),
                  up = pmax(.data$doc.x, .data$doc.y)) %>%
    dplyr::distinct(.data$up, .data$dn) %>%
    dplyr::select(a = dn, b = up) %>%
    dplyr::arrange(.data$a, .data$b) %>%
    dplyr::mutate(score = NA_real_)

  class(candidates) <- c("textreuse_candidates", class(candidates))

  candidates

}

#' Recompute the hashes for a document or corpus
#'
#' Given a \code{\link{TextReuseTextDocument}} or a
#' \code{\link{TextReuseCorpus}}, this function recomputes either the hashes or
#' the minhashes with the function specified. This implies that you have
#' retained the tokens with the \code{keep_tokens = TRUE} parameter.
#'
#' @param x A \code{\link{TextReuseTextDocument}} or
#'   \code{\link{TextReuseCorpus}}.
#' @param func A function to either hash the tokens or to generate the minhash
#'   signature. See \code{\link{hash_string}}, \code{\link{minhash_generator}}.
#' @param type Recompute the \code{hashes} or \code{minhashes}?
#'
#' @return The modified \code{\link{TextReuseTextDocument}} or
#'   \code{\link{TextReuseCorpus}}.
#'
#' @examples
#' dir <- system.file("extdata/legal", package = "textreuse")
#' minhash1 <- minhash_generator(seed = 1)
#' corpus <- TextReuseCorpus(dir = dir, minhash_func = minhash1, keep_tokens = TRUE)
#' head(minhashes(corpus[[1]]))
#' minhash2 <- minhash_generator(seed = 2)
#' corpus <- rehash(corpus, minhash2, type = "minhashes")
#' head(minhashes(corpus[[2]]))
#'
#' @export
rehash <- function(x, func, type = c("hashes", "minhashes")) {
  UseMethod("rehash", x)
}

#' @export
rehash.TextReuseTextDocument <- function(x, func,
                                         type = c("hashes", "minhashes")) {
  assert_that(has_tokens(x),
              is.function(func))
  type <- match.arg(type)

  if (type == "hashes") {
    x$hashes <- func(x$tokens)
    x$meta$hash_func <- as.character(substitute(func))
  } else if (type == "minhashes") {
    x$minhashes <- func(x$tokens)
    x$meta$minhash_func <- as.character(substitute(func))
  }
  x
}

#' @export
rehash.TextReuseCorpus <- function(x, func,  type = c("hashes", "minhashes")) {
  assert_that(is.function(func))
  type <- match.arg(type)
  apply_func <- get_apply_function()
  x$documents <- apply_func(x$documents, rehash, func, type)
  if (type == "hashes")
    x$meta$hash_func <- as.character(substitute(func))
  else if (type == "minhashes")
    x$meta$minhash_func <- as.character(substitute(func))
  x
}

#' Recompute the tokens for a document or corpus
#'
#' Given a \code{\link{TextReuseTextDocument}} or a
#' \code{\link{TextReuseCorpus}}, this function recomputes the tokens and hashes
#' with the functions specified. Optionally, it can also recompute the minhash signatures.
#'
#' @param x A \code{\link{TextReuseTextDocument}} or
#'   \code{\link{TextReuseCorpus}}.
#' @param tokenizer A function to split the text into tokens. See
#'   \code{\link{tokenizers}}.
#' @param ... Arguments passed on to the \code{tokenizer}.
#' @param hash_func A function to hash the tokens. See
#'   \code{\link{hash_string}}.
#' @param minhash_func A function to create minhash signatures. See
#'   \code{\link{minhash_generator}}.
#' @param keep_tokens Should the tokens be saved in the document that is
#'   returned or discarded?
#' @param keep_text Should the text be saved in the document that is returned or
#'   discarded?
#'
#' @return The modified \code{\link{TextReuseTextDocument}} or
#'   \code{\link{TextReuseCorpus}}.
#'
#' @examples
#' dir <- system.file("extdata/legal", package = "textreuse")
#' corpus <- TextReuseCorpus(dir = dir, tokenizer = NULL)
#' corpus <- tokenize(corpus, tokenize_ngrams)
#' head(tokens(corpus[[1]]))
#' @export
tokenize <- function(x, tokenizer, ..., hash_func = hash_string,
                     minhash_func = NULL, keep_tokens = FALSE,
                     keep_text = TRUE) {
  UseMethod("tokenize", x)
}

#' @export
tokenize.TextReuseTextDocument <- function(x, tokenizer, ...,
                                           hash_func = hash_string,
                                           minhash_func = NULL,
                                           keep_tokens = TRUE,
                                           keep_text = TRUE) {
  assert_that(has_content(x),
              is.function(tokenizer),
              is.function(hash_func))
  x$tokens <- tokenizer(x$content, ...)
  x$hashes <- hash_func(x$tokens)
  if (!keep_tokens) x$tokens <- NULL
  if (!keep_text) x$text <- NULL
  x$meta$tokenizer <- as.character(substitute(tokenizer))
  x$meta$hash_func <- as.character(substitute(hash_func))
  if (!is.null(minhash_func)) {
    x$minhash <- minhash_func(x$tokens)
    x$meta$minhash_func <- as.character(substitute(minhash_func))
  } else {
    # If tokens are redone, minhashes are invalid, so delete them if they are
    # not also recomputed.
    x$minhashes <- NULL
    x$meta$minhash_func <- NULL
  }
  x
}

#' @export
tokenize.TextReuseCorpus <- function(x, tokenizer, ..., hash_func = hash_string,
                                     minhash_func = NULL, keep_tokens = TRUE,
                                     keep_text = TRUE) {
  apply_func <- get_apply_function()
  x$documents <- apply_func(x$documents, tokenize, tokenizer, ...,
                            hash_func = hash_func, minhash_func = minhash_func,
                            keep_tokens = keep_tokens, keep_text = keep_text)

  x$meta$tokenizer <- as.character(substitute(tokenizer))
  x$meta$hash_func <- as.character(substitute(hash_func))
  if (!is.null(minhash_func)) {
    x$meta$minhash_func <- as.character(substitute(minhash_func))
  } else {
    x$meta$minhash_func <- NULL
  }

  x

}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' Hash a string to an integer
#' @param x A character vector to be hashed.
#' @return A vector of integer hashes.
#' @examples
#' s <- c("How", "many", "roads", "must", "a", "man", "walk", "down")
#' hash_string(s)
#' @export
hash_string <- function(x) {
    .Call(`_textreuse_hash_string`, x)
}

shingle_ngrams <- function(words, n) {
    .Call(`_textreuse_shingle_ngrams`, words, n)
}

skip_ngrams <- function(words, n, k) {
    .Call(`_textreuse_skip_ngrams`, words, n, k)
}

sw_matrix <- function(m, a, b, match, mismatch, gap, progress) {
    .Call(`_textreuse_sw_matrix`, m, a, b, match, mismatch, gap, progress)
}

context("Hashing")

lines  <- system.file("extdata/legal/ny1850-match.txt", package = "textreuse") %>%
  readLines()
ngrams <- lines %>%
  str_c(collapse = " ") %>%
  tokenize_ngrams(n = 5)

lines_hashed  <- hash_string(lines)
ngrams_hashed <- hash_string(ngrams)

test_that("returns correct values", {
  expect_is(lines_hashed, "integer")
  expect_is(ngrams_hashed, "integer")
})

test_that("there are no collisions", {
  expect_equal(length(unique(lines)), length(unique(lines_hashed)))
  expect_equal(length(unique(ngrams)), length(unique(ngrams_hashed)))
})

#' Split texts into tokens
#'
#' These functions each turn a text into tokens. The \code{tokenize_ngrams}
#' functions returns shingled n-grams.
#'
#' @name tokenizers
#' @param string A character vector of length 1 to be tokenized.
#' @param lowercase Should the tokens be made lower case?
#' @param n For n-gram tokenizers, the number of words in each n-gram.
#' @param k For the skip n-gram tokenizer, the maximum skip distance between
#'   words. The function will compute all skip n-grams between \code{0} and
#'   \code{k}.
#' @details These functions will strip all punctuation.
#' @return A character vector containing the tokens.
#' @examples
#' dylan <- "How many roads must a man walk down? The answer is blowin' in the wind."
#' tokenize_words(dylan)
#' tokenize_sentences(dylan)
#' tokenize_ngrams(dylan, n = 2)
#' tokenize_skip_ngrams(dylan, n = 3, k = 2)
NULL

#' @export
#' @rdname tokenizers
tokenize_words <- function(string, lowercase = TRUE) {
  assert_that(assertthat::is.string(string))
  out <- str_split(string, boundary("word"))[[1]]
  if (lowercase) str_to_lower(out) else out
}

#' @export
#' @rdname tokenizers
tokenize_sentences <- function(string, lowercase = TRUE) {
  assert_that(assertthat::is.string(string))
  out <- str_split(string, boundary("sentence", skip_word_none = FALSE))[[1]]
  out <- str_replace_all(out, "[[:punct:]]", " ")
  out <- str_replace_all(out, "\\s+", " ")
  out <- str_trim(out)
  if (lowercase) str_to_lower(out) else out
}

#' @export
#' @rdname tokenizers
tokenize_ngrams <- function(string, lowercase = TRUE, n = 3) {
  assert_that(is.count(n),
              assertthat::is.string(string))
  words <- tokenize_words(string, lowercase = lowercase)
  assert_that(n < length(words))
  shingle_ngrams(words, n = n)
}

#' @export
#' @rdname tokenizers
tokenize_skip_ngrams <- function(string, lowercase = TRUE, n = 3, k = 1) {
  assert_that(is.count(n),
              is.count(k) | k == 0,
              assertthat::is.string(string))
  words <- tokenize_words(string, lowercase = lowercase)
  assert_that(n + n * k - k <= length(words))
  skip_ngrams(words, n = n, k = k)
}

context("Tokenizers")

sentence <- "This is a sentence which has a number of words in it; also some
             tricky puncuation and spacing. Does it work?"

test_that("n-grams can be generated", {
  results <- c("this is a sentence which",
               "is a sentence which has",
               "a sentence which has a",
               "sentence which has a number",
               "which has a number of",
               "has a number of words",
               "a number of words in",
               "number of words in it",
               "of words in it also",
               "words in it also some",
               "in it also some tricky",
               "it also some tricky puncuation",
               "also some tricky puncuation and",
               "some tricky puncuation and spacing",
               "tricky puncuation and spacing does",
               "puncuation and spacing does it",
               "and spacing does it work")
  expect_equal(tokenize_ngrams(sentence, n = 5), results)
})

test_that("different values of n work", {
  n3 <- tokenize_ngrams(sentence, n = 3)
  n5 <- tokenize_ngrams(sentence, n = 5)

  expect_equal(str_split(n3[1], boundary("word"))[[1]] %>% length(), 3)
  expect_equal(str_split(n5[1], boundary("word"))[[1]] %>% length(), 5)
})

test_that("ngrams can be uppercase", {
  sentence <- tokenize_ngrams("This is a Capital Word.", n = 4, lowercase = FALSE)
  results  <- c("This is a Capital", "is a Capital Word")
  expect_equal(sentence, results)
})

test_that("word tokenizer works", {
  expect_equal(tokenize_words("There are several; WORDS here"),
               c("there", "are", "several", "words", "here"))
  expect_equal(tokenize_words("There are several; WORDS here", lowercase = FALSE),
               c("There", "are", "several", "WORDS", "here"))
})


test_that("sentence tokenizers works", {
  expect_equal(tokenize_sentences("This is a---sentence. This too."),
               c("this is a sentence", "this too"))
  expect_equal(tokenize_sentences("This is a sentence. This too.", lowercase = FALSE),
               c("This is a sentence", "This too"))
})

test_that("tokenizers fail on non-strings", {
  text <- c("This is not a string", "because it is length two.")
  expect_error(tokenize_words(text), "string is not a string")
  expect_error(tokenize_sentences(text), "string is not a string")
  expect_error(tokenize_ngrams(text), "string is not a string")
  expect_error(tokenize_skip_ngrams(text), "string is not a string")
})

test_that("skip n-grams behave as expected", {
  dylan <- "How many roads must a man walk down"
  skips <- tokenize_skip_ngrams(dylan, n = 3, k = 2)
  correct <- c("how many roads", "many roads must", "roads must a", "must a man",
               "a man walk", "man walk down", "how must walk", "many a down",
               "how roads a", "many must man", "roads a walk", "must man down")
  expect_equal(sort(skips), sort(correct))
  expect_equal(tokenize_skip_ngrams(dylan, n = 5, k = 0),
               tokenize_ngrams(dylan, n = 5))
})

#' List of all candidates in a corpus
#'
#' @param candidates A data frame of candidate pairs from
#'   \code{\link{lsh_candidates}}.
#' @return A character vector of document IDs from the candidate pairs, to be
#'   used to subset the \code{\link{TextReuseCorpus}}.
#' @examples
#' dir <- system.file("extdata/legal", package = "textreuse")
#' minhash <- minhash_generator(200, seed = 234)
#' corpus <- TextReuseCorpus(dir = dir,
#'                           tokenizer = tokenize_ngrams, n = 5,
#'                           minhash_func = minhash)
#' buckets <- lsh(corpus, bands = 50)
#' candidates <- lsh_candidates(buckets)
#' lsh_subset(candidates)
#' corpus[lsh_subset(candidates)]
#' @export
lsh_subset <- function(candidates) {
  assert_that(is_candidates_df(candidates))
  sort(unique(c(candidates$a, candidates$b)))
}

#' Local alignment of natural language texts
#'
#' This function takes two texts, either as strings or as
#' \code{TextReuseTextDocument} objects, and finds the optimal local alignment
#' of those texts. A local alignment finds the best matching subset of the two
#' documents. This function adapts the
#' \href{https://en.wikipedia.org/wiki/Smith-Waterman_algorithm}{Smith-Waterman
#' algorithm}, used for genetic sequencing, for use with natural language. It
#' compare the texts word by word (the comparison is case-insensitive) and
#' scores them according to a set of parameters. These parameters define the
#' score for a \code{match}, and the penalties for a \code{mismatch} and for
#' opening a \code{gap} (i.e., the first mismatch in a potential sequence). The
#' function then reports the optimal local alignment. Only the subset of the
#' documents that is a match is included. Insertions or deletions in the text
#' are reported with the \code{edit_mark} character.
#'
#' @param a A character vector of length one, or a
#'   \code{\link{TextReuseTextDocument}}.
#' @param b A character vector of length one, or a
#'   \code{\link{TextReuseTextDocument}}.
#' @param match The score to assign a matching word. Should be a positive
#'   integer.
#' @param mismatch The score to assign a mismatching word. Should be a negative
#'   integer or zero.
#' @param gap The penalty for opening a gap in the sequence. Should be a
#'   negative integer or zero.
#' @param edit_mark A single character used for displaying for displaying
#'   insertions/deletions in the documents.
#' @param preserve_punctuation Preserve punctuation in the displayed alignment.
#'   The alignment still compares tokens after stripping punctuation.
#' @param progress Display a progress bar and messages while computing the
#'   alignment.
#'
#' @return A list with the class \code{textreuse_alignment}. This list contains
#'   several elements: \itemize{ \item \code{a_edit} and \code{b_edit}:
#'   Character vectors of the sequences with edits marked. \item \code{score}:
#'   The score of the optimal alignment. }
#'
#' @details
#'
#' The compute time of this function is proportional to the product of the
#' lengths of the two documents. Thus, longer documents will take considerably
#' more time to compute. This function has been tested with pairs of documents
#' containing about 25 thousand words each.
#'
#' If the function reports that there were multiple optimal alignments, then it
#' is likely that there is no strong match in the document.
#'
#' The score reported for the local alignment is dependent on both the size of
#' the documents and on the strength of the match, as well as on the parameters
#' for match, mismatch, and gap penalties, so the scores are not directly
#' comparable.
#'
#' @references For a useful description of the algorithm, see
#'   \href{http://etherealbits.com/2013/04/string-alignment-dynamic-programming-dna/}{this
#'   post}. For the application of the Smith-Waterman algorithm to natural
#'   language, see David A. Smith, Ryan Cordell, and Elizabeth Maddock Dillon,
#'   "Infectious Texts: Modeling Text Reuse in Nineteenth-Century Newspapers,"
#'   IEEE International Conference on Big Data, 2013.
#'
#' @examples
#' align_local("The answer is blowin' in the wind.",
#'             "As the Bob Dylan song says, the answer is blowing in the wind.")
#'
#' # Example of matching documents from a corpus
#' dir <- system.file("extdata/legal", package = "textreuse")
#' corpus <- TextReuseCorpus(dir = dir, progress = FALSE)
#' alignment <- align_local(corpus[["ca1851-match"]], corpus[["ny1850-match"]])
#' str(alignment)
#'
#' @export
align_local <- function(a, b, match = 2L, mismatch = -1L, gap = -1L,
                        edit_mark = "#", preserve_punctuation = FALSE,
                        progress = interactive()) {
 assert_that(identical(class(a), class(b)))
 UseMethod("align_local", a)
}

#' @export
align_local.TextReuseTextDocument <- function(a, b, match = 2L, mismatch = -1L,
                                              gap = -1L, edit_mark = "#",
                                              preserve_punctuation = FALSE,
                                              progress = interactive()) {
  align_local(content(a), content(b), match = match, mismatch = mismatch,
              gap = gap, edit_mark = edit_mark,
              preserve_punctuation = preserve_punctuation)
}

#' @export
align_local.default <- function(a, b, match = 2L, mismatch = -1L, gap = -1L,
                                edit_mark = "#", preserve_punctuation = FALSE,
                                progress = interactive()) {

  assert_that(is.string(a),
              is.string(b),
              is_integer_like(match),
              is_integer_like(mismatch),
              is_integer_like(gap),
              is.string(edit_mark),
              is.flag(preserve_punctuation))

  if (match <= 0 || mismatch > 0 || gap > 0 || !(str_length(edit_mark) == 1)) {
    stop("The scoring parameters should have the following characteristics:\n",
         "    - `match` should be a positive integer\n",
         "    - `mismatch` should be a negative integer or zero\n",
         "    - `gap` should be a negative integer or zero\n",
         "    - `edit_mark` should be a single character\n")
  }

  # Keep everything as integers because IntegerMatrix saves memory
  match    <- as.integer(match)
  mismatch <- as.integer(mismatch)
  gap      <- as.integer(gap)

  # Prepare the character vectors. Tokenize to words to compare word by word.
  # Use all lower case for the comparison, but use original capitalization in
  # the output.
  a_orig <- align_tokens(a, preserve_punctuation = preserve_punctuation)
  b_orig <- align_tokens(b, preserve_punctuation = preserve_punctuation)
  a <- normalize_alignment_tokens(a_orig)
  b <- normalize_alignment_tokens(b_orig)

  # Only show a progress bar for long computations
  n_rows <- length(b) + 1
  n_cols <- length(a) + 1
  if (n_rows * n_cols < 1e7) progress <- FALSE

  # Create the integer matrix
  if (progress) {
    message("Preparing a matrix with ",
            prettyNum(n_rows * n_cols, big.mark = ","),
            " elements.")
  }
  m <- matrix(0L, n_rows, n_cols)

  # Calculate the matrix of possible paths
  if (progress) message("Computing the optimal local alignment.")
  m <- sw_matrix(m, a, b, match, mismatch, gap, progress)

  # Find the starting place in the matrix
  alignment_score <- max(m)
  if (alignment_score == 0) {
    alignment <- list(a_edits = "", b_edits = "", score = alignment_score)
    class(alignment) <- c("textreuse_alignment", "list")
    return(alignment)
  }

  max_match <- which(m == alignment_score, arr.ind = TRUE, useNames = FALSE)

  if (nrow(max_match) > 1) {
    warning("Multiple optimal local alignments found; selecting only one of them.",
            call. = FALSE)
  }

  if (progress) message("Extracting the local alignment.")

  # Create output vectors which are as long as conceivably necessary
  a_out <- vector(mode = "character", length = max(max_match))
  b_out <- vector(mode = "character", length = max(max_match))
  a_out[] <- NA_character_
  b_out[] <- NA_character_

  # Initialize counters for the matrix and the output vector
  row_i <- max_match[1, 1]
  col_i <- max_match[1, 2]
  out_i <- 1L

  # Place our first known values in the output vectors
  b_out[out_i] <- b_orig[row_i - 1]
  a_out[out_i] <- a_orig[col_i - 1]
  out_i = out_i + 1L # Advance the out vector position

  # Begin moving up, left, or diagonally within the matrix till we hit a zero
  while (m[row_i - 1, col_i - 1] != 0) {

    # Values of the current cell, the cells up, left, diagonal, and the max
    up       <- m[row_i - 1, col_i]
    left     <- m[row_i, col_i - 1]
    diagn    <- m[row_i - 1, col_i - 1]
    max_cell <- max(up, left, diagn)

    # Move in the direction of the maximum cell. If there are ties, choose up
    # first, then left, then diagonal. Privilege up and left because they
    # preserve edits.
    #
    # In each case add the current words to the out vectors. For moves up and
    # and left there will be an insertion/deletion, so add a symbol like ####
    # that is the same number of characters as the word in the other vector.
    #
    # Note that the index of the matrix is offset by one from character vectors
    # a and b, so we use the row and column indices - 1. The column corresponds
    # to `a` and the rows correspond to `b`.
    if (up == max_cell) {
      row_i <- row_i - 1
      bword <- b_orig[row_i - 1]
      b_out[out_i] <- bword
      a_out[out_i] <- mark_chars(bword, edit_mark)
    } else if (left == max_cell) {
      col_i <- col_i - 1
      aword <- a_orig[col_i - 1]
      b_out[out_i] <- mark_chars(aword, edit_mark)
      a_out[out_i] <- aword
    } else if (diagn == max_cell) {
      row_i <- row_i - 1
      col_i <- col_i - 1
      bword <-  b_orig[row_i - 1]
      aword <- a_orig[col_i - 1]
      # Diagonals are a special case, because instead of an insertion or a
      # deletion we might have a substitution of words. If that is the case,
      # then treat it like a double insertion and deletion.
      if (a[col_i - 1] == b[row_i - 1]) {
        b_out[out_i] <- bword
        a_out[out_i] <- aword
      } else {
        b_out[out_i] <- bword
        a_out[out_i] <- mark_chars(bword, edit_mark)
        out_i <- out_i + 1
        b_out[out_i] <- mark_chars(aword, edit_mark)
        a_out[out_i] <- aword
      }
    }

    # Move forward one position in the out vectors, no matter which direction
    # we moved
    out_i <- out_i + 1

  }

  # Clean up the outputs
  b_out <- str_c(rev(b_out[!is.na(b_out)]), collapse = " ")
  a_out <- str_c(rev(a_out[!is.na(a_out)]), collapse = " ")

  # Create the alignment object
  alignment <- list(a_edits = a_out, b_edits = b_out, score = alignment_score)
  class(alignment) <- c("textreuse_alignment", "list")

  alignment

}

align_tokens <- function(x, preserve_punctuation) {
  if (!preserve_punctuation) return(tokenize_words(x, lowercase = FALSE))
  tokens <- str_split(str_squish(x), "\\s+")[[1]]
  tokens[tokens != ""]
}

normalize_alignment_tokens <- function(x) {
  str_to_lower(str_replace_all(x, "[[:punct:]]", ""))
}

#' @export
print.textreuse_alignment <- function(x, ...) {
  cat("TextReuse alignment\n")
  cat("Alignment score:", x$score, "\n")
  cat("Document A:\n")
  cat(str_wrap(x$a_edits, width = 72))
  cat("\n\nDocument B:\n")
  cat(str_wrap(x$b_edits, width = 72))
  cat("\n\n")
  invisible(x)
}

#' Pairwise comparisons among documents in a corpus
#'
#' Given a \code{\link{TextReuseCorpus}} containing documents of class
#' \code{\link{TextReuseTextDocument}}, this function applies a comparison
#' function to every pairing of documents, and returns a matrix with the
#' comparison scores.
#'
#' @param corpus A \code{\link{TextReuseCorpus}}.
#' @param f The function to apply to \code{x} and \code{y}.
#' @param ... Additional arguments passed to \code{f}.
#' @param directional Some comparison functions are commutative, so that
#'   \code{f(a, b) == f(b, a)} (e.g., \code{\link{jaccard_similarity}}). Other
#'   functions are directional, so that \code{f(a, b)} measures \code{a}'s
#'   borrowing from \code{b}, which may not be the same as \code{f(b, a)} (e.g.,
#'   \code{\link{ratio_of_matches}}). If \code{directional} is \code{FALSE},
#'   then only the minimum number of comparisons will be made, i.e., the upper
#'   triangle of the matrix. If \code{directional} is \code{TRUE}, then both
#'   directional comparisons will be measured. In no case, however, will
#'   documents be compared to themselves, i.e., the diagonal of the matrix.
#' @param progress Display a progress bar while comparing documents.
#'
#' @return A square matrix with dimensions equal to the length of the corpus,
#'   and row and column names set by the names of the documents in the corpus. A
#'   value of \code{NA} in the matrix indicates that a comparison was not made.
#'   In cases of directional comparisons, then the comparison reported is
#'   \code{f(row, column)}.
#'
#' @seealso See these document comparison functions,
#'   \code{\link{jaccard_similarity}}, \code{\link{ratio_of_matches}}.
#'
#' @examples
#' dir <- system.file("extdata/legal", package = "textreuse")
#' corpus <- TextReuseCorpus(dir = dir)
#' names(corpus) <- filenames(names(corpus))
#'
#' # A non-directional comparison
#' pairwise_compare(corpus, jaccard_similarity)
#'
#' # A directional comparison
#' pairwise_compare(corpus, ratio_of_matches, directional = TRUE)
#' @export
pairwise_compare <- function(corpus, f, ..., directional = FALSE,
                        progress = interactive()) {
  assert_that(is.TextReuseCorpus(corpus),
              is.function(f))

  len <- length(corpus)
  ids <- names(corpus)

  m <- matrix(0, len, len, dimnames = list(ids, ids))

  if (!directional)
    m[lower.tri(m, diag = TRUE)] <- NA
  else
    diag(m) <- NA


  if (progress) {
    num_pairs <- sum(!is.na(m))
    message("Making ", prettyNum(num_pairs, big.mark = ","), " comparisons.")
    pb <- txtProgressBar(min = 0, max = num_pairs, style = 3)
  }

  for (i in seq_along(m)) {
    if (is.na(m[i])) next
    indexes <- arrayInd(i, dim(m))
    m[indexes] <- f(corpus[[indexes[1]]], corpus[[indexes[2]]])
    if (progress) setTxtProgressBar(pb, getTxtProgressBar(pb) + 1)
  }

  if (progress) close(pb)

  m

}

#' Build an index of tokens and documents
#'
#' Build an inverted index from tokens to the documents that contain them. This
#' is useful for finding document pairs that share one or more n-grams without
#' comparing every document pair. The corpus must be created with
#' \code{keep_tokens = TRUE}.
#'
#' @param corpus A \code{\link{TextReuseCorpus}} with retained tokens.
#' @param min_doc_count Minimum number of documents a token must appear in to
#'   be retained. Increase this to remove rare tokens.
#' @param max_doc_count Maximum number of documents a token may appear in to be
#'   retained. Decrease this to remove very common tokens.
#' @return A \code{textreuse_token_index} data frame with columns \code{token},
#'   \code{docs}, and \code{n_docs}.
#' @export
token_index <- function(corpus, min_doc_count = 2, max_doc_count = Inf) {
  assert_that(is.TextReuseCorpus(corpus),
              is.count(min_doc_count),
              is.number(max_doc_count),
              all(vapply(tokens(corpus), Negate(is.null), logical(1))))

  entries <- lapply(names(corpus), function(doc_id) {
    tibble::tibble(token = unique(tokens(corpus[[doc_id]])), doc = doc_id)
  })

  index <- dplyr::bind_rows(entries) %>%
    dplyr::group_by(.data$token) %>%
    dplyr::summarize(docs = list(sort(.data$doc)),
                     n_docs = dplyr::n(),
                     .groups = "drop") %>%
    dplyr::filter(.data$n_docs >= min_doc_count,
                  .data$n_docs <= max_doc_count) %>%
    dplyr::arrange(.data$token)

  class(index) <- c("textreuse_token_index", class(index))

  index
}

#' Extract candidate document pairs from a token index
#'
#' @param index A \code{textreuse_token_index} object returned by
#'   \code{\link{token_index}}.
#' @return A \code{textreuse_candidates} data frame.
#' @export
token_index_candidates <- function(index) {
  assert_that(inherits(index, "textreuse_token_index"))

  pair_matrices <- lapply(index$docs, function(doc_ids) {
    if (length(doc_ids) < 2) return(NULL)
    t(utils::combn(doc_ids, 2))
  })
  pair_matrices <- Filter(Negate(is.null), pair_matrices)

  if (length(pair_matrices) == 0) {
    candidates <- tibble::tibble(a = character(), b = character(),
                                 score = numeric())
  } else {
    candidates <- do.call(rbind, pair_matrices) %>%
      as.data.frame(stringsAsFactors = FALSE) %>%
      tibble::as_tibble() %>%
      stats::setNames(c("a", "b")) %>%
      dplyr::mutate(a = pmin(.data$a, .data$b),
                    b = pmax(.data$a, .data$b)) %>%
      dplyr::distinct(.data$a, .data$b) %>%
      dplyr::arrange(.data$a, .data$b) %>%
      dplyr::mutate(score = NA_real_)
  }

  class(candidates) <- c("textreuse_candidates", class(candidates))

  candidates
}

context("Token index")

texts <- c(a = "one two three four",
           b = "one two three five",
           c = "six seven eight nine")
corpus <- TextReuseCorpus(text = texts, tokenizer = tokenize_ngrams, n = 2,
                          keep_tokens = TRUE)

test_that("token index maps tokens to documents", {
  index <- token_index(corpus, min_doc_count = 2)
  expect_is(index, "textreuse_token_index")
  expect_equal(index$token, c("one two", "two three"))
  expect_equal(index$n_docs, c(2L, 2L))
  expect_equal(index$docs[[1]], c("a", "b"))
})

test_that("token index can remove common tokens", {
  index <- token_index(corpus, min_doc_count = 2, max_doc_count = 1)
  expect_equal(nrow(index), 0)
})

test_that("token index can return candidate pairs", {
  index <- token_index(corpus, min_doc_count = 2)
  candidates <- token_index_candidates(index)
  expect_is(candidates, "textreuse_candidates")
  expect_equal(candidates$a, "a")
  expect_equal(candidates$b, "b")
  expect_equal(candidates$score, NA_real_)
})

#' Count words
#'
#' This function counts words in a text, for example, a character vector, a
#' \code{\link{TextReuseTextDocument}}, some other object that inherits from
#' \code{\link[NLP]{TextDocument}}, or a all the documents in a
#' \code{\link{TextReuseCorpus}}.
#'
#' @param x The object containing a text.
#' @export
#' @return An integer vector for the word count.
wordcount <- function(x) UseMethod("wordcount", x)

#' @export
wordcount.default <- function(x) {
  assert_that(is.string(x))
  str_count(x, boundary("word"))
}

#' @export
wordcount.TextDocument <- function(x) wordcount(x$content)

#' @export
wordcount.TextReuseCorpus <- function(x) {
  vapply(x$documents, wordcount, integer(1))
}

---
title: "Pairwise comparisons for document similarity"
author:
  - "Lincoln Mullen"
  - "Yaoxiang Li"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Pairwise comparisons for document similarity}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

The most straightforward way to compare documents within a corpus is to compare each document to every other document. 

First we will load the corpus and tokenize it with shingled n-grams.

```{r}
library(textreuse)
dir <- system.file("extdata/ats", package = "textreuse")
corpus <- TextReuseCorpus(dir = dir, tokenizer = tokenize_ngrams, n = 5,
                          progress = FALSE)
```

We can use any of the comparison functions to compare two documents in the corpus. (Note that these functions, when applied to documents, compare their hashed tokens and not the tokens directly.)

```{r}
jaccard_similarity(corpus[["remember00palm"]], 
                   corpus[["remembermeorholy00palm"]])
```

The `pairwise_compare()` function applies a comparison function to each pair of documents in a corpus. The result is a matrix with the scores for each comparison.

```{r eval=FALSE}
comparisons <- pairwise_compare(corpus, jaccard_similarity, progress = FALSE)
comparisons[1:4, 1:4]
```

```{r, echo=FALSE}
comparisons <- pairwise_compare(corpus, jaccard_similarity, progress = FALSE)
round(comparisons[1:3, 1:3], digits = 3)
```

If you prefer, you can convert the matrix of all comparisons to a data frame of pairs and scores. Here we create the data frame and keep only the pairs with scores above a significant value.

```{r}
candidates <- pairwise_candidates(comparisons)
candidates[candidates$score > 0.1, ]
```

The pairwise comparison method is inadequate for a corpus of any size, however. For a corpus of size $n$, the number of comparisons (assuming the comparisons are commutative) is $\frac{n^2 - n}{2}$. A corpus of 100 documents would require 4,950 comparisons; a corpus of 1,000 documents would require 499,500 comparisons. A better approach for corpora of any appreciable size is to use the minhash/LSH algorithms described in another vignette: 

```{r eval=FALSE}
vignette("minhash", package = "textreuse")
```

---
title: "Text Alignment"
author:
  - "Lincoln Mullen"
  - "Yaoxiang Li"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Text Alignment}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

Local alignment is the process of taking two documents and finding the best subset of each document that aligns with one another. A commonly used local alignment algorithm for genetics is the [Smith-Waterman algorithm](https://en.wikipedia.org/wiki/Smith-Waterman_algorithm). This package offers a version of the Smith-Waterman algorithm intended to be used for natural language processing.

Consider these two documents. The first is part of Shakespeare's *Measure for Measure*. The second is a made-up piece of literary criticism quoting the play, but our imaginary literary critic has bungled the quotation. This is a common class of problems (not bungling literary critics but) documents which contain pieces, often heavily modified, from other documents.

```{r}
shakespeare <- paste(
  "Haste still pays haste, and leisure answers leisure;",
  "Like doth quit like, and MEASURE still FOR MEASURE.",
  "Then, Angelo, thy fault's thus manifested;",
  "Which, though thou wouldst deny, denies thee vantage.",
  "We do condemn thee to the very block",
  "Where Claudio stoop'd to death, and with like haste.",
  "Away with him!")
critic <- paste(
  "The play comes to its culmination where Duke Vincentio, quoting from",
  "the words of the Sermon on the Mount, says,",
  "'Haste still goes very quickly , and leisure answers leisure;",
  "Like doth cancel like, and measure still for measure.'",
  "These titular words sum up the meaning of the play.")
```

We can uses the local alignment function to extract the part of the text that was borrowed. Notice that the resulting object shows us the changes that have been made.

```{r}
library(textreuse)
align_local(shakespeare, critic)
```

See the documentation for the function to see how to tune the match: `?align_local`. This function works with character vectors or with documents of class `TextReuseTextDocument`.

---
title: "Minhash and locality-sensitive hashing"
author:
  - "Lincoln Mullen"
  - "Yaoxiang Li"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Minhash and locality-sensitive hashing}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, echo=FALSE, message=FALSE}
library("dplyr")
```

Performing pairwise comparisons in a corpus is time-consuming because the number of comparisons grows geometrically with the size of the corpus. Most of those comparisons, furthermore, are unnecessary because they do not result in matches. The combination of minhash and locality-sensitive hashing (LSH) seeks to solve these problems. They make it possible to compute possible matches only once for each document, so that the cost of computation grows linearly rather than exponentially. This vignette explains how to use the minhash and locality-sensitive hashing functions in this package. For an explanation of why they work, see Jure Leskovec, Anand Rajaraman, and Jeff Ullman, *Mining of Massive Datasets* (Cambridge University Press, 2011), ch. 3. (This [blog post](https://matthewcasperson.blogspot.com/2013/11/minhash-for-dummies.html) is a more succinct explanation.)

We begin by creating a minhash function. A minhash function converts tokenized text into a set of hash integers, then selects the minimum value. This is the equivalent of randomly selecting a token. The function then does the same thing repeatedly with different hashing functions, in effect selecting `n` random shingles. The additional hashing functions come from a bitwise XOR with random integers. That is why the `minhash_generator()` accepts a seed, so that we can re-create the same minhash function again. In other words, a minhash function converts a set of tokens of any length into `n` randomly selected and hashed tokens.

```{r}
library(textreuse)
minhash <- minhash_generator(n = 240, seed = 3552)
head(minhash(c("turn tokens into", "tokens into hashes", "into hashes fast")))
```

Now when we load our corpus, we will tokenize our texts as usual, but we will use our generated `minhash()` function to compute the hashes. We specify that we want to create a minhash signature by passing our minhash function to the `minhash_func =` parameter.

```{r}
dir <- system.file("extdata/ats", package = "textreuse")
corpus <- TextReuseCorpus(dir = dir, tokenizer = tokenize_ngrams, n = 5,
                          minhash_func = minhash, keep_tokens = TRUE,
                          progress = FALSE)
```

We can verify that we have minhashes in our corpus:

```{r}
head(minhashes(corpus[[1]]))
length(minhashes(corpus[[1]]))
```


Now all our documents are represented by `n = 240` randomly selected and hashed shingles. Comparing those shingles should be the equivalent of finding the Jaccard similarity of the two documents. However, we still have the problem of pairwise comparison.

The locality-sensitive hashing algorithm, provided in this package by the `lsh()` function, solves this problem. LSH breaks the minhashes into a series of bands comprised of rows. For example, 200 minhashes might broken into 50 bands of 4 rows each. Each band is hashed to a bucket. If two documents have the exact same minhashes in a band, they will be hashed to the same bucket, and so will be considered candidate pairs. Each pair of documents has as many chances to be considered a candidate as their are bands, and the fewer rows there are in each band, the more likely it is that each document will match another.

How likely is it, then, that we will detect a match? The probability of a match depends on the Jaccard similarity of a pair of documents. The more similar two documents are, the more likely they are to be considered candidates, which is what we want. The probability of a match is an S-curve (see Leskovec, Rajaraman, and Ullman), so there is a threshold Jaccard similarity above which documents are likely to be a match. We can calculate the likely threshold based on the number of minhashes and bands that we are using.

```{r}
lsh_threshold(h = 200, b = 50)
lsh_threshold(h = 240, b = 80)
```

Using 240 minhashes and 80 bands, we will likely detect documents with an actual Jaccard similarity of above 0.232. We can also estimate the probability that a pair of documents with a Jaccard similarity `s` will be marked as potential matches.

```{r}
lsh_probability(h = 240, b = 80, s = 0.25)
lsh_probability(h = 240, b =  80, s = 0.75)
```

These numbers seem reasonable for our purposes, so we will set the number of minhashes at 240 and the number of bands at 80.

Now we can use the `lsh()` function to calculate the locality-sensitive hashes for our documents. 

```{r}
buckets <- lsh(corpus, bands = 80, progress = FALSE)
buckets
```

Note that using the LSH method only requires us to calculate the signatures (or buckets) for each document one time. This implies that we can take several data frames of LSH signatures and bind their rows together (e.g., with `dplyr::bind_rows()`). This permits us to compute the signatures for only part of a corpus at a time, or to continue to add to the corpus. Note, however, that you **must** use the same minhash function, generating the same number of minhashes and using the same seed and you **must** use the same number of bands in order to get valid results.

We can extract the potential matches from the cache using `lsh_query()` or  `lsh_candidates()`. The first function returns matches for only one document, specified by its ID; the second functions returns all potential pairs of matches.

```{r}
baxter_matches <- lsh_query(buckets, "calltounconv00baxt")
baxter_matches
candidates <- lsh_candidates(buckets)
candidates
```

Notice that LSH has identified the same three pairs of documents as potential matches that we found with pairwise comparisons, but did so much faster. But we do not have similarity scores; we only know that these documents are likely to have Jaccard similarity scores above the `r round(lsh_threshold(h = 240, b = 80), 3)` threshold.

Now we can use `lsh_compare()` to apply a similarity function to the candidate pairs of documents. Note that we only have to do 3 comparisons for all the candidates, instead of 28 pairs when comparing all 8 documents in the corpus pairwise.

```{r}
lsh_compare(candidates, corpus, jaccard_similarity, progress = FALSE)
```

Note that these results are identical to what we calculated in the pairwise vignette, but required much less computation.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/minhash.R
\name{minhash_generator}
\alias{minhash_generator}
\title{Generate a minhash function}
\usage{
minhash_generator(n = 200, seed = NULL)
}
\arguments{
\item{n}{The number of minhashes that the returned function should generate.}

\item{seed}{An option parameter to set the seed used in generating the random
numbers to ensure that the same minhash function is used on repeated
applications.}
}
\value{
A function which will take a character vector and return \code{n}
  minhashes.
}
\description{
A minhash value is calculated by hashing the strings in a character vector to
integers and then selecting the minimum value. Repeated minhash values are
generated by using different hash functions: these different hash functions
are created by using performing a bitwise \code{XOR} operation
(\code{\link{bitwXor}}) with a vector of random integers. Since it is vital
that the same random integers be used for each document, this function
generates another function which will always use the same integers. The
returned function is intended to be passed to the \code{hash_func} parameter
of \code{\link{TextReuseTextDocument}}.
}
\examples{
set.seed(253)
minhash <- minhash_generator(10)

# Example with a TextReuseTextDocument
file <- system.file("extdata/legal/ny1850-match.txt", package = "textreuse")
doc <- TextReuseTextDocument(file = file, hash_func = minhash,
                             keep_tokens = TRUE)
hashes(doc)

# Example with a character vector
is.character(tokens(doc))
minhash(tokens(doc))
}
\references{
Jure Leskovec, Anand Rajaraman, and Jeff Ullman,
  \emph{Mining of Massive Datasets} (Cambridge University Press, 2011), ch.
  3. See also Matthew Casperson,
  "\href{http://matthewcasperson.blogspot.com/2013/11/minhash-for-dummies.html}{Minhash
   for Dummies}" (November 14, 2013).
}
\seealso{
\code{\link{lsh}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filenames.R
\name{filenames}
\alias{filenames}
\title{Filenames from paths}
\usage{
filenames(paths, extension = FALSE)
}
\arguments{
\item{paths}{A character vector of paths.}

\item{extension}{Should the file extension be preserved?}
}
\description{
This function takes a character vector of paths and returns just the file
name, by default without the extension. A \code{\link{TextReuseCorpus}} uses
the paths to the files in the corpus as the names of the list. This function
is intended to turn those paths into more manageable identifiers.
}
\examples{
paths <- c("corpus/one.txt", "corpus/two.md", "corpus/three.text")
filenames(paths)
filenames(paths, extension = TRUE)
}
\seealso{
\code{\link{basename}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/token_index.R
\name{token_index}
\alias{token_index}
\title{Build an index of tokens and documents}
\usage{
token_index(corpus, min_doc_count = 2, max_doc_count = Inf)
}
\arguments{
\item{corpus}{A \code{\link{TextReuseCorpus}} with retained tokens.}

\item{min_doc_count}{Minimum number of documents a token must appear in to
be retained. Increase this to remove rare tokens.}

\item{max_doc_count}{Maximum number of documents a token may appear in to be
retained. Decrease this to remove very common tokens.}
}
\value{
A \code{textreuse_token_index} data frame with columns \code{token},
  \code{docs}, and \code{n_docs}.
}
\description{
Build an inverted index from tokens to the documents that contain them. This
is useful for finding document pairs that share one or more n-grams without
comparing every document pair. The corpus must be created with
\code{keep_tokens = TRUE}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TextReuseTextDocument.R
\name{TextReuseTextDocument}
\alias{TextReuseTextDocument}
\alias{is.TextReuseTextDocument}
\alias{has_content}
\alias{has_tokens}
\alias{has_hashes}
\alias{has_minhashes}
\title{TextReuseTextDocument}
\usage{
TextReuseTextDocument(
  text,
  file = NULL,
  meta = list(),
  tokenizer = tokenize_ngrams,
  ...,
  hash_func = hash_string,
  minhash_func = NULL,
  keep_tokens = FALSE,
  keep_text = TRUE,
  skip_short = TRUE,
  encoding = "unknown"
)

is.TextReuseTextDocument(x)

has_content(x)

has_tokens(x)

has_hashes(x)

has_minhashes(x)
}
\arguments{
\item{text}{A character vector containing the text of the document. This
argument can be skipped if supplying \code{file}.}

\item{file}{The path to a text file, if \code{text} is not provided.}

\item{meta}{A list with named elements for the metadata associated with this
document. If a document is created using the \code{text} parameter, then
you must provide an \code{id} field, e.g., \code{meta = list(id =
"my_id")}. If the document is created using \code{file}, then the ID will
be created from the file name.}

\item{tokenizer}{A function to split the text into tokens. See
\code{\link{tokenizers}}. If value is \code{NULL}, then tokenizing and
hashing will be skipped.}

\item{...}{Arguments passed on to the \code{tokenizer}.}

\item{hash_func}{A function to hash the tokens. See
\code{\link{hash_string}}.}

\item{minhash_func}{A function to create minhash signatures of the document.
See \code{\link{minhash_generator}}.}

\item{keep_tokens}{Should the tokens be saved in the document that is
returned or discarded?}

\item{keep_text}{Should the text be saved in the document that is returned or
discarded?}

\item{skip_short}{Should short documents be skipped? (See details.)}

\item{encoding}{Encoding to be used when reading files.}

\item{x}{An R object to check.}
}
\value{
An object of class \code{TextReuseTextDocument}. This object inherits
  from the virtual S3 class \code{\link[NLP]{TextDocument}} in the NLP
  package. It contains the following elements: \describe{ \item{content}{The
  text of the document.} \item{tokens}{The tokens created from the text.}
  \item{hashes}{Hashes created from the tokens.} \item{minhashes}{The minhash
  signature of the document.} \item{metadata}{The document metadata,
  including the filename (if any) in \code{file}.} }
}
\description{
This is the constructor function for \code{TextReuseTextDocument} objects.
This class is used for comparing documents.
}
\details{
This constructor function follows a three-step process. It reads in
  the text, either from a file or from memory. It then tokenizes that text.
  Then it hashes the tokens. Most of the comparison functions in this package
  rely only on the hashes to make the comparison. By passing \code{FALSE} to
  \code{keep_tokens} and \code{keep_text}, you can avoid saving those
  objects, which can result in significant memory savings for large corpora.

  If \code{skip_short = TRUE}, this function will return \code{NULL} for very
  short or empty documents. A very short document is one where there are too
  few words to create at least two n-grams. For example, if five-grams are
  desired, then a document must be at least six words long. If no value of
  \code{n} is provided, then the function assumes a value of \code{n = 3}. A
  warning will be printed with the document ID of a skipped document.
}
\examples{
file <- system.file("extdata/legal/ny1850-match.txt", package = "textreuse")
doc  <- TextReuseTextDocument(file = file, meta = list(id = "ny1850"))
print(doc)
meta(doc)
head(tokens(doc))
head(hashes(doc))
\dontrun{
content(doc)
}
}
\seealso{
\link[=TextReuseTextDocument-accessors]{Accessors for TextReuse
  objects}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lsh_query.R
\name{lsh_query}
\alias{lsh_query}
\title{Query a LSH cache for matches to a single document}
\usage{
lsh_query(buckets, id)
}
\arguments{
\item{buckets}{An \code{lsh_buckets} object created by \code{\link{lsh}}.}

\item{id}{The document ID to find matches for.}
}
\value{
An \code{lsh_candidates} data frame with matches to the document specified.
}
\description{
This function retrieves the matches for a single document from an \code{lsh_buckets} object created by \code{\link{lsh}}. See \code{\link{lsh_candidates}} to retrieve all pairs of matches.
}
\examples{
dir <- system.file("extdata/legal", package = "textreuse")
minhash <- minhash_generator(200, seed = 235)
corpus <- TextReuseCorpus(dir = dir,
                          tokenizer = tokenize_ngrams, n = 5,
                          minhash_func = minhash)
buckets <- lsh(corpus, bands = 50)
lsh_query(buckets, "ny1850-match")

}
\seealso{
\code{\link{lsh}}, \code{\link{lsh_candidates}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/textreuse-package.R
\docType{package}
\name{textreuse-package}
\alias{textreuse}
\alias{textreuse-package}
\title{textreuse: Detect Text Reuse and Document Similarity}
\description{
Tools for measuring similarity among documents and detecting passages which have been reused. Implements shingled n-gram, skip n-gram, and other tokenizers; similarity/dissimilarity functions; pairwise comparisons; minhash and locality sensitive hashing algorithms; and a version of the Smith-Waterman local alignment algorithm suitable for natural language.
}
\details{
The best place to begin with this package in the introductory vignette.

\code{vignette("textreuse-introduction", package = "textreuse")}

After reading that vignette, the "pairwise" and "minhash" vignettes introduce
specific paths for working with the package.

\code{vignette("textreuse-pairwise", package = "textreuse")}

\code{vignette("textreuse-minhash", package = "textreuse")}

\code{vignette("textreuse-alignment", package = "textreuse")}

Another good place to begin with the package is the documentation for loading
documents (\code{\link{TextReuseTextDocument}} and
\code{\link{TextReuseCorpus}}), for \link{tokenizers},
\link[=similarity-functions]{similarity functions}, and
\link[=lsh]{locality-sensitive hashing}.
}
\references{
The sample data provided in the \code{extdata/ats} directory
  contains nineteenth-century American Tract Society publications gathered
  from the \href{https://archive.org/}{Internet Archive}.

  The sample data provided in the \code{extdata/legal} directory, are taken
  from the following nineteenth-century codes of civil procedure from
  California and New York.

  \emph{Final Report of the Commissioners on Practice and Pleadings}, in 2
  \emph{Documents of the Assembly of New York}, 73rd Sess., No. 16, (1850):
  243-250, sections 597-613.
  \href{http://books.google.com/books?id=9HEbAQAAIAAJ&pg=PA243#v=onepage&q&f=false}{Google
   Books}.

  \emph{An Act To Regulate Proceedings in Civil Cases}, 1851 \emph{California
  Laws} 51, 51-53 sections 4-17; 101, sections 313-316.
  \href{http://books.google.com/books?id=4PHEAAAAIAAJ&pg=PA51#v=onepage&q&f=false}{Google
   Books}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://docs.ropensci.org/textreuse/}
  \item \url{https://github.com/ropensci/textreuse}
  \item Report bugs at \url{https://github.com/ropensci/textreuse/issues}
}

}
\author{
\strong{Maintainer}: Yaoxiang Li \email{liyaoxiang@outlook.com} (\href{https://orcid.org/0000-0001-9200-1016}{ORCID})

Authors:
\itemize{
  \item Lincoln Mullen \email{lincoln@lincolnmullen.com} (\href{https://orcid.org/0000-0001-5103-6917}{ORCID})
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/similarity.R
\name{similarity-functions}
\alias{similarity-functions}
\alias{jaccard_similarity}
\alias{jaccard_dissimilarity}
\alias{jaccard_bag_similarity}
\alias{ratio_of_matches}
\alias{count_matches}
\alias{matching_tokens}
\title{Measure similarity/dissimilarity in documents}
\usage{
jaccard_similarity(a, b)

jaccard_dissimilarity(a, b)

jaccard_bag_similarity(a, b)

ratio_of_matches(a, b)

count_matches(a, b)

matching_tokens(a, b)
}
\arguments{
\item{a}{The first set (or bag) to be compared. The origin bag for
directional comparisons.}

\item{b}{The second set (or bag) to be compared. The destination bag for
directional comparisons.}
}
\description{
A set of functions which take two sets or bag of words and measure their
similarity or dissimilarity.
}
\details{
The functions \code{jaccard_similarity} and
  \code{jaccard_dissimilarity} provide the Jaccard measures of similarity or
  dissimilarity for two sets. The coefficients will be numbers between
  \code{0} and \code{1}. For the similarity coefficient, the higher the
  number the more similar the two sets are. When applied to two documents of
  class \code{\link{TextReuseTextDocument}}, the hashes in those documents
  are compared. But this function can be passed objects of any class accepted
  by the set functions in base R. So it is possible, for instance, to pass
  this function two character vectors comprised of word, line, sentence, or
  paragraph tokens, or those character vectors hashed as integers.

  The Jaccard similarity coeffecient is defined as follows:

  \deqn{J(A, B) = \frac{ | A \cap B | }{ | A \cup B | }}{ length(intersect(a,
  b)) / length(union(a, b))}

  The Jaccard dissimilarity is simply

  \deqn{1 - J(A, B)}

  The function \code{jaccard_bag_similarity} treats \code{a} and \code{b} as
  bags rather than sets, so that the result is a fraction where the numerator
  is the sum of each matching element counted the minimum number of times it
  appears in each bag, and the denominator is the sum of the lengths of both
  bags. The maximum value for the Jaccard bag similarity is \code{0.5}.

  The function \code{ratio_of_matches} finds the ratio between the number of
  items in \code{b} that are also in \code{a} and the total number of items
  in \code{b}. Note that this similarity measure is directional: it measures
  how much \code{b} borrows from \code{a}, but says nothing about how much of
  \code{a} borrows from \code{b}.

  The function \code{count_matches} returns the numerator used by
  \code{ratio_of_matches}: the number of items in \code{b} also found in
  \code{a}. The function \code{matching_tokens} returns those matching items
  from \code{b}, preserving their order and duplicates.
}
\examples{
jaccard_similarity(1:6, 3:10)
jaccard_dissimilarity(1:6, 3:10)

a <- c("a", "a", "a", "b")
b <- c("a", "a", "b", "b", "c")
jaccard_similarity(a, b)
jaccard_bag_similarity(a, b)
ratio_of_matches(a, b)
ratio_of_matches(b, a)
count_matches(a, b)
matching_tokens(a, b)

ny         <- system.file("extdata/legal/ny1850-match.txt", package = "textreuse")
ca_match   <- system.file("extdata/legal/ca1851-match.txt", package = "textreuse")
ca_nomatch <- system.file("extdata/legal/ca1851-nomatch.txt", package = "textreuse")

ny         <- TextReuseTextDocument(file = ny,
                                    meta = list(id = "ny"))
ca_match   <- TextReuseTextDocument(file = ca_match,
                                    meta = list(id = "ca_match"))
ca_nomatch <- TextReuseTextDocument(file = ca_nomatch,
                                    meta = list(id = "ca_nomatch"))

# These two should have higher similarity scores
jaccard_similarity(ny, ca_match)
ratio_of_matches(ny, ca_match)

# These two should have lower similarity scores
jaccard_similarity(ny, ca_nomatch)
ratio_of_matches(ny, ca_nomatch)

}
\references{
Jure Leskovec, Anand Rajaraman, and Jeff Ullman,
  \emph{Mining of Massive Datasets} (Cambridge University Press, 2011).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lsh.R
\name{lsh}
\alias{lsh}
\title{Locality sensitive hashing for minhash}
\usage{
lsh(x, bands, progress = interactive())
}
\arguments{
\item{x}{A \code{\link{TextReuseCorpus}} or
\code{\link{TextReuseTextDocument}}.}

\item{bands}{The number of bands to use for locality sensitive hashing. The
number of hashes in the documents in the corpus must be evenly divisible by
the number of bands. See \code{\link{lsh_threshold}} and
\code{\link{lsh_probability}} for guidance in selecting the number of bands
and hashes.}

\item{progress}{Display a progress bar while comparing documents.}
}
\value{
A data frame (with the additional class \code{lsh_buckets}),
 containing a column with the document IDs and a column with their LSH
 signatures, or buckets.
}
\description{
Locality sensitive hashing (LSH) discovers potential matches among a corpus of
documents quickly, so that only likely pairs can be compared.
}
\details{
Locality sensitive hashing is a technique for detecting document
 similarity that does not require pairwise comparisons. When comparing pairs
 of documents, the number of pairs grows rapidly, so that only the smallest
 corpora can be compared pairwise in a reasonable amount of computation time.
 Locality sensitive hashing, on the other hand, takes a document which has
 been tokenized and hashed using a minhash algorithm. (See
 \code{\link{minhash_generator}}.) Each set of minhash signatures is then
 broken into bands comprised of a certain number of rows. (For example, 200
 minhash signatures might be broken down into 20 bands each containing 10
 rows.) Each band is then hashed to a bucket. Documents with identical rows
 in a band will be hashed to the same bucket. The likelihood that a document
 will be marked as a potential duplicate is proportional to the number of
 bands and inversely proportional to the number of rows in each band.

 This function returns a data frame with the additional class
 \code{lsh_buckets}. The LSH technique only requires that the signatures for
 each document be calculated once. So it is possible, as long as one uses the
 same minhash function and the same number of bands, to combine the outputs
 from this function at different times. The output can thus be treated as a
 kind of cache of LSH signatures.

 To extract pairs of documents from the output of this function, see
 \code{\link{lsh_candidates}}.
}
\examples{
dir <- system.file("extdata/legal", package = "textreuse")
minhash <- minhash_generator(200, seed = 235)
corpus <- TextReuseCorpus(dir = dir,
                          tokenizer = tokenize_ngrams, n = 5,
                          minhash_func = minhash)
buckets <- lsh(corpus, bands = 50)
buckets
}
\references{
Jure Leskovec, Anand Rajaraman, and Jeff Ullman,
 \emph{Mining of Massive Datasets} (Cambridge University Press, 2011), ch.
 3. See also Matthew Casperson,
 "\href{http://matthewcasperson.blogspot.com/2013/11/minhash-for-dummies.html}{Minhash
  for Dummies}" (November 14, 2013).
}
\seealso{
\code{\link{minhash_generator}}, \code{\link{lsh_add}},
 \code{\link{lsh_candidates}}, \code{\link{lsh_query}},
 \code{\link{lsh_probability}},
 \code{\link{lsh_threshold}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pairwise_candidates.R
\name{pairwise_candidates}
\alias{pairwise_candidates}
\title{Candidate pairs from pairwise comparisons}
\usage{
pairwise_candidates(m, directional = FALSE)
}
\arguments{
\item{m}{A matrix from \code{\link{pairwise_compare}}.}

\item{directional}{Should be set to the same value as in
\code{\link{pairwise_compare}}.}
}
\value{
A data frame containing all the non-\code{NA} values from \code{m}.
  Columns \code{a} and \code{b} are the IDs from the original corpus as
  passed to the comparison function. Column \code{score} is the score
  returned by the comparison function.
}
\description{
Converts a comparison matrix generated by \code{\link{pairwise_compare}} into a
data frame of candidates for matches.
}
\examples{
dir <- system.file("extdata/legal", package = "textreuse")
corpus <- TextReuseCorpus(dir = dir)

m1 <- pairwise_compare(corpus, ratio_of_matches, directional = TRUE)
pairwise_candidates(m1, directional = TRUE)

m2 <- pairwise_compare(corpus, jaccard_similarity)
pairwise_candidates(m2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wordcount.R
\name{wordcount}
\alias{wordcount}
\title{Count words}
\usage{
wordcount(x)
}
\arguments{
\item{x}{The object containing a text.}
}
\value{
An integer vector for the word count.
}
\description{
This function counts words in a text, for example, a character vector, a
\code{\link{TextReuseTextDocument}}, some other object that inherits from
\code{\link[NLP]{TextDocument}}, or a all the documents in a
\code{\link{TextReuseCorpus}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{hash_string}
\alias{hash_string}
\title{Hash a string to an integer}
\usage{
hash_string(x)
}
\arguments{
\item{x}{A character vector to be hashed.}
}
\value{
A vector of integer hashes.
}
\description{
Hash a string to an integer
}
\examples{
s <- c("How", "many", "roads", "must", "a", "man", "walk", "down")
hash_string(s)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lsh_subset.R
\name{lsh_subset}
\alias{lsh_subset}
\title{List of all candidates in a corpus}
\usage{
lsh_subset(candidates)
}
\arguments{
\item{candidates}{A data frame of candidate pairs from
\code{\link{lsh_candidates}}.}
}
\value{
A character vector of document IDs from the candidate pairs, to be
  used to subset the \code{\link{TextReuseCorpus}}.
}
\description{
List of all candidates in a corpus
}
\examples{
dir <- system.file("extdata/legal", package = "textreuse")
minhash <- minhash_generator(200, seed = 234)
corpus <- TextReuseCorpus(dir = dir,
                          tokenizer = tokenize_ngrams, n = 5,
                          minhash_func = minhash)
buckets <- lsh(corpus, bands = 50)
candidates <- lsh_candidates(buckets)
lsh_subset(candidates)
corpus[lsh_subset(candidates)]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tokenizers.R
\name{tokenizers}
\alias{tokenizers}
\alias{tokenize_words}
\alias{tokenize_sentences}
\alias{tokenize_ngrams}
\alias{tokenize_skip_ngrams}
\title{Split texts into tokens}
\usage{
tokenize_words(string, lowercase = TRUE)

tokenize_sentences(string, lowercase = TRUE)

tokenize_ngrams(string, lowercase = TRUE, n = 3)

tokenize_skip_ngrams(string, lowercase = TRUE, n = 3, k = 1)
}
\arguments{
\item{string}{A character vector of length 1 to be tokenized.}

\item{lowercase}{Should the tokens be made lower case?}

\item{n}{For n-gram tokenizers, the number of words in each n-gram.}

\item{k}{For the skip n-gram tokenizer, the maximum skip distance between
words. The function will compute all skip n-grams between \code{0} and
\code{k}.}
}
\value{
A character vector containing the tokens.
}
\description{
These functions each turn a text into tokens. The \code{tokenize_ngrams}
functions returns shingled n-grams.
}
\details{
These functions will strip all punctuation.
}
\examples{
dylan <- "How many roads must a man walk down? The answer is blowin' in the wind."
tokenize_words(dylan)
tokenize_sentences(dylan)
tokenize_ngrams(dylan, n = 2)
tokenize_skip_ngrams(dylan, n = 3, k = 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rehash.R
\name{rehash}
\alias{rehash}
\title{Recompute the hashes for a document or corpus}
\usage{
rehash(x, func, type = c("hashes", "minhashes"))
}
\arguments{
\item{x}{A \code{\link{TextReuseTextDocument}} or
\code{\link{TextReuseCorpus}}.}

\item{func}{A function to either hash the tokens or to generate the minhash
signature. See \code{\link{hash_string}}, \code{\link{minhash_generator}}.}

\item{type}{Recompute the \code{hashes} or \code{minhashes}?}
}
\value{
The modified \code{\link{TextReuseTextDocument}} or
  \code{\link{TextReuseCorpus}}.
}
\description{
Given a \code{\link{TextReuseTextDocument}} or a
\code{\link{TextReuseCorpus}}, this function recomputes either the hashes or
the minhashes with the function specified. This implies that you have
retained the tokens with the \code{keep_tokens = TRUE} parameter.
}
\examples{
dir <- system.file("extdata/legal", package = "textreuse")
minhash1 <- minhash_generator(seed = 1)
corpus <- TextReuseCorpus(dir = dir, minhash_func = minhash1, keep_tokens = TRUE)
head(minhashes(corpus[[1]]))
minhash2 <- minhash_generator(seed = 2)
corpus <- rehash(corpus, minhash2, type = "minhashes")
head(minhashes(corpus[[2]]))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/align_local.R
\name{align_local}
\alias{align_local}
\title{Local alignment of natural language texts}
\usage{
align_local(
  a,
  b,
  match = 2L,
  mismatch = -1L,
  gap = -1L,
  edit_mark = "#",
  preserve_punctuation = FALSE,
  progress = interactive()
)
}
\arguments{
\item{a}{A character vector of length one, or a
\code{\link{TextReuseTextDocument}}.}

\item{b}{A character vector of length one, or a
\code{\link{TextReuseTextDocument}}.}

\item{match}{The score to assign a matching word. Should be a positive
integer.}

\item{mismatch}{The score to assign a mismatching word. Should be a negative
integer or zero.}

\item{gap}{The penalty for opening a gap in the sequence. Should be a
negative integer or zero.}

\item{edit_mark}{A single character used for displaying for displaying
insertions/deletions in the documents.}

\item{preserve_punctuation}{Preserve punctuation in the displayed alignment.
The alignment still compares tokens after stripping punctuation.}

\item{progress}{Display a progress bar and messages while computing the
alignment.}
}
\value{
A list with the class \code{textreuse_alignment}. This list contains
  several elements: \itemize{ \item \code{a_edit} and \code{b_edit}:
  Character vectors of the sequences with edits marked. \item \code{score}:
  The score of the optimal alignment. }
}
\description{
This function takes two texts, either as strings or as
\code{TextReuseTextDocument} objects, and finds the optimal local alignment
of those texts. A local alignment finds the best matching subset of the two
documents. This function adapts the
\href{https://en.wikipedia.org/wiki/Smith-Waterman_algorithm}{Smith-Waterman
algorithm}, used for genetic sequencing, for use with natural language. It
compare the texts word by word (the comparison is case-insensitive) and
scores them according to a set of parameters. These parameters define the
score for a \code{match}, and the penalties for a \code{mismatch} and for
opening a \code{gap} (i.e., the first mismatch in a potential sequence). The
function then reports the optimal local alignment. Only the subset of the
documents that is a match is included. Insertions or deletions in the text
are reported with the \code{edit_mark} character.
}
\details{
The compute time of this function is proportional to the product of the
lengths of the two documents. Thus, longer documents will take considerably
more time to compute. This function has been tested with pairs of documents
containing about 25 thousand words each.

If the function reports that there were multiple optimal alignments, then it
is likely that there is no strong match in the document.

The score reported for the local alignment is dependent on both the size of
the documents and on the strength of the match, as well as on the parameters
for match, mismatch, and gap penalties, so the scores are not directly
comparable.
}
\examples{
align_local("The answer is blowin' in the wind.",
            "As the Bob Dylan song says, the answer is blowing in the wind.")

# Example of matching documents from a corpus
dir <- system.file("extdata/legal", package = "textreuse")
corpus <- TextReuseCorpus(dir = dir, progress = FALSE)
alignment <- align_local(corpus[["ca1851-match"]], corpus[["ny1850-match"]])
str(alignment)

}
\references{
For a useful description of the algorithm, see
  \href{http://etherealbits.com/2013/04/string-alignment-dynamic-programming-dna/}{this
  post}. For the application of the Smith-Waterman algorithm to natural
  language, see David A. Smith, Ryan Cordell, and Elizabeth Maddock Dillon,
  "Infectious Texts: Modeling Text Reuse in Nineteenth-Century Newspapers,"
  IEEE International Conference on Big Data, 2013.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lsh.R
\name{lsh_add}
\alias{lsh_add}
\title{Add documents to a LSH cache}
\usage{
lsh_add(buckets, x, bands, progress = interactive())
}
\arguments{
\item{buckets}{An \code{lsh_buckets} object created by \code{\link{lsh}}.}

\item{x}{A \code{\link{TextReuseCorpus}} or
\code{\link{TextReuseTextDocument}} with minhashes.}

\item{bands}{The number of bands to use for locality sensitive hashing. The
number of hashes in the documents in the corpus must be evenly divisible by
the number of bands. See \code{\link{lsh_threshold}} and
\code{\link{lsh_probability}} for guidance in selecting the number of bands
and hashes.}

\item{progress}{Display a progress bar while comparing documents.}
}
\value{
An updated \code{lsh_buckets} object.
}
\description{
This function adds buckets for one or more new documents to an existing
\code{lsh_buckets} object. Use the same \code{bands} value and minhash
function that were used to create the original buckets.
}
\seealso{
\code{\link{lsh}}, \code{\link{lsh_query}},
  \code{\link{lsh_candidates}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lsh_candidates.R
\name{lsh_candidates}
\alias{lsh_candidates}
\title{Candidate pairs from LSH comparisons}
\usage{
lsh_candidates(buckets)
}
\arguments{
\item{buckets}{A data frame returned from \code{\link{lsh}}.}
}
\value{
A data frame of candidate pairs.
}
\description{
Given a data frame of LSH buckets returned from \code{\link{lsh}}, this
function returns the potential candidates.
}
\examples{
dir <- system.file("extdata/legal", package = "textreuse")
minhash <- minhash_generator(200, seed = 234)
corpus <- TextReuseCorpus(dir = dir,
                          tokenizer = tokenize_ngrams, n = 5,
                          minhash_func = minhash)
buckets <- lsh(corpus, bands = 50)
lsh_candidates(buckets)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TextReuseCorpus.R
\name{TextReuseCorpus}
\alias{TextReuseCorpus}
\alias{is.TextReuseCorpus}
\alias{skipped}
\title{TextReuseCorpus}
\usage{
TextReuseCorpus(
  paths,
  dir = NULL,
  text = NULL,
  meta = list(),
  progress = interactive(),
  tokenizer = tokenize_ngrams,
  ...,
  hash_func = hash_string,
  minhash_func = NULL,
  keep_tokens = FALSE,
  keep_text = TRUE,
  skip_short = TRUE,
  encoding = "unknown"
)

is.TextReuseCorpus(x)

skipped(x)
}
\arguments{
\item{paths}{A character vector of paths to files to be opened.}

\item{dir}{The path to a directory of text files.}

\item{text}{A character vector (possibly named) of documents.}

\item{meta}{A list with named elements for the metadata associated with this
corpus.}

\item{progress}{Display a progress bar while loading files.}

\item{tokenizer}{A function to split the text into tokens. See
\code{\link{tokenizers}}. If value is \code{NULL}, then tokenizing and
hashing will be skipped.}

\item{...}{Arguments passed on to the \code{tokenizer}.}

\item{hash_func}{A function to hash the tokens. See
\code{\link{hash_string}}.}

\item{minhash_func}{A function to create minhash signatures of the document.
See \code{\link{minhash_generator}}.}

\item{keep_tokens}{Should the tokens be saved in the documents that are
returned or discarded?}

\item{keep_text}{Should the text be saved in the documents that are returned
or discarded?}

\item{skip_short}{Should short documents be skipped? (See details.)}

\item{encoding}{Encoding to be used when reading files.}

\item{x}{An R object to check.}
}
\description{
This is the constructor function for a \code{TextReuseCorpus}, modeled on the
virtual S3 class \code{Corpus} from the \code{tm} package. The
object is a \code{TextReuseCorpus}, which is basically a list containing
objects of class \code{\link{TextReuseTextDocument}}. Arguments are passed
along to that constructor function. To create the corpus, you can pass either
a character vector of paths to text files using the \code{paths =} parameter,
a directory containing text files (with any extension) using the \code{dir =}
parameter, or a character vector of documents using the \code{text = }
parameter, where each element in the characer vector is a document. If the
character vector passed to \code{text = } has names, then those names will be
used as the document IDs. Otherwise, IDs will be assigned to the documents.
Only one of the \code{paths}, \code{dir}, or \code{text} parameters should be
specified.
}
\details{
If \code{skip_short = TRUE}, this function will skip very short or
  empty documents. A very short document is one where there are too few words
  to create at least two n-grams. For example, if five-grams are desired,
  then a document must be at least six words long. If no value of \code{n} is
  provided, then the function assumes a value of \code{n = 3}. A warning will
  be printed with the document ID of each skipped document. Use
  \code{skipped()} to get the IDs of skipped documents.

  This function will use multiple cores on non-Windows machines if the
  \code{"mc.cores"} option is set. For example, to use four cores:
  \code{options("mc.cores" = 4L)}.
}
\examples{
dir <- system.file("extdata/legal", package = "textreuse")
corpus <- TextReuseCorpus(dir = dir, meta = list("description" = "Field Codes"))
# Subset by position or file name
corpus[[1]]
names(corpus)
corpus[["ca1851-match"]]

}
\seealso{
\link[=TextReuseTextDocument-accessors]{Accessors for TextReuse
  objects}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TextReuseTextDocument.R
\name{TextReuseTextDocument-accessors}
\alias{TextReuseTextDocument-accessors}
\alias{tokens}
\alias{tokens<-}
\alias{hashes}
\alias{hashes<-}
\alias{minhashes}
\alias{minhashes<-}
\title{Accessors for TextReuse objects}
\usage{
tokens(x)

tokens(x) <- value

hashes(x)

hashes(x) <- value

minhashes(x)

minhashes(x) <- value
}
\arguments{
\item{x}{The object to access.}

\item{value}{The value to assign.}
}
\value{
Either a vector or a named list of vectors.
}
\description{
Accessor functions to read and write components of
\code{\link{TextReuseTextDocument}} and \code{\link{TextReuseCorpus}}
objects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lsh_probability.R
\name{lsh_probability}
\alias{lsh_probability}
\alias{lsh_threshold}
\title{Probability that a candidate pair will be detected with LSH}
\usage{
lsh_probability(h, b, s)

lsh_threshold(h, b)
}
\arguments{
\item{h}{The number of minhash signatures.}

\item{b}{The number of LSH bands.}

\item{s}{The Jaccard similarity.}
}
\description{
Functions to help choose the correct parameters for the \code{\link{lsh}} and
\code{\link{minhash_generator}} functions. Use \code{lsh_threshold} to
determine the minimum Jaccard similarity for two documents for them to likely
be considered a match. Use \code{lsh_probability} to determine the
probability that a pair of documents with a known Jaccard similarity will be
detected.
}
\details{
Locality sensitive hashing returns a list of possible matches for
similar documents. How likely is it that a pair of documents will be detected
as a possible match? If \code{h} is the number of minhash signatures,
\code{b} is the number of bands in the LSH function (implying then that the
number of rows \code{r = h / b}), and \code{s} is the actual Jaccard
similarity of the two documents, then the probability \code{p} that the two
documents will be marked as a candidate pair is given by this equation.

\deqn{p = 1 - (1 - s^{r})^{b}}

According to \href{http://infolab.stanford.edu/~ullman/mmds/book.pdf}{MMDS},
that equation approximates an S-curve. This implies that there is a threshold
(\code{t}) for \code{s} approximated by this equation.

\deqn{t = \frac{1}{b}^{\frac{1}{r}}}
}
\examples{
# Threshold for default values
lsh_threshold(h = 200, b = 40)

# Probability for varying values of s
lsh_probability(h = 200, b = 40, s = .25)
lsh_probability(h = 200, b = 40, s = .50)
lsh_probability(h = 200, b = 40, s = .75)
}
\references{
Jure Leskovec, Anand Rajaraman, and Jeff Ullman,
 \emph{Mining of Massive Datasets} (Cambridge University Press, 2011), ch. 3.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/token_index.R
\name{token_index_candidates}
\alias{token_index_candidates}
\title{Extract candidate document pairs from a token index}
\usage{
token_index_candidates(index)
}
\arguments{
\item{index}{A \code{textreuse_token_index} object returned by
\code{\link{token_index}}.}
}
\value{
A \code{textreuse_candidates} data frame.
}
\description{
Extract candidate document pairs from a token index
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lsh_compare.R
\name{lsh_compare}
\alias{lsh_compare}
\title{Compare candidates identified by LSH}
\usage{
lsh_compare(candidates, corpus, f, progress = interactive())
}
\arguments{
\item{candidates}{A data frame returned by \code{\link{lsh_candidates}}.}

\item{corpus}{The same \code{\link{TextReuseCorpus}} corpus which was used to generate the candidates.}

\item{f}{A comparison function such as \code{\link{jaccard_similarity}}.}

\item{progress}{Display a progress bar while comparing documents. Progress
bars are disabled when using parallel processing.}
}
\value{
A data frame with values calculated for \code{score}.
}
\description{
The \code{\link{lsh_candidates}} only identifies potential matches, but
cannot estimate the actual similarity of the documents. This function takes a
data frame returned by \code{\link{lsh_candidates}} and applies a comparison
function to each of the documents in a corpus, thereby calculating the
document similarity score. Note that since your corpus will have minhash
signatures rather than hashes for the tokens itself, you will probably wish
to use \code{\link{tokenize}} to calculate new hashes. This can be done for
just the potentially similar documents. See the package vignettes for
details.
}
\examples{
dir <- system.file("extdata/legal", package = "textreuse")
minhash <- minhash_generator(200, seed = 234)
corpus <- TextReuseCorpus(dir = dir,
                          tokenizer = tokenize_ngrams, n = 5,
                          minhash_func = minhash)
buckets <- lsh(corpus, bands = 50)
candidates <- lsh_candidates(buckets)
lsh_compare(candidates, corpus, jaccard_similarity)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TextReuseTextDocument.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{meta}
\alias{meta<-}
\alias{content}
\alias{content<-}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{NLP}{\code{\link[NLP:generics]{content}}, \code{\link[NLP:generics]{content<-}}, \code{\link[NLP:generics]{meta}}, \code{\link[NLP:generics]{meta<-}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conversion-functions.R
\name{as.matrix.textreuse_candidates}
\alias{as.matrix.textreuse_candidates}
\alias{as_sparse_matrix}
\title{Convert candidates data frames to other formats}
\usage{
\method{as.matrix}{textreuse_candidates}(x, ...)

as_sparse_matrix(x)
}
\arguments{
\item{x}{An object of class \code{\link[=lsh_compare]{textreuse_candidates}}.}

\item{...}{Additional arguments.}
}
\value{
A similarity matrix with row and column names containing document IDs.
}
\description{
These functions convert a \code{textreuse_candidates} object to dense or
sparse matrices.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pairwise_compare.R
\name{pairwise_compare}
\alias{pairwise_compare}
\title{Pairwise comparisons among documents in a corpus}
\usage{
pairwise_compare(corpus, f, ..., directional = FALSE, progress = interactive())
}
\arguments{
\item{corpus}{A \code{\link{TextReuseCorpus}}.}

\item{f}{The function to apply to \code{x} and \code{y}.}

\item{...}{Additional arguments passed to \code{f}.}

\item{directional}{Some comparison functions are commutative, so that
\code{f(a, b) == f(b, a)} (e.g., \code{\link{jaccard_similarity}}). Other
functions are directional, so that \code{f(a, b)} measures \code{a}'s
borrowing from \code{b}, which may not be the same as \code{f(b, a)} (e.g.,
\code{\link{ratio_of_matches}}). If \code{directional} is \code{FALSE},
then only the minimum number of comparisons will be made, i.e., the upper
triangle of the matrix. If \code{directional} is \code{TRUE}, then both
directional comparisons will be measured. In no case, however, will
documents be compared to themselves, i.e., the diagonal of the matrix.}

\item{progress}{Display a progress bar while comparing documents.}
}
\value{
A square matrix with dimensions equal to the length of the corpus,
  and row and column names set by the names of the documents in the corpus. A
  value of \code{NA} in the matrix indicates that a comparison was not made.
  In cases of directional comparisons, then the comparison reported is
  \code{f(row, column)}.
}
\description{
Given a \code{\link{TextReuseCorpus}} containing documents of class
\code{\link{TextReuseTextDocument}}, this function applies a comparison
function to every pairing of documents, and returns a matrix with the
comparison scores.
}
\examples{
dir <- system.file("extdata/legal", package = "textreuse")
corpus <- TextReuseCorpus(dir = dir)
names(corpus) <- filenames(names(corpus))

# A non-directional comparison
pairwise_compare(corpus, jaccard_similarity)

# A directional comparison
pairwise_compare(corpus, ratio_of_matches, directional = TRUE)
}
\seealso{
See these document comparison functions,
  \code{\link{jaccard_similarity}}, \code{\link{ratio_of_matches}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tokenize.R
\name{tokenize}
\alias{tokenize}
\title{Recompute the tokens for a document or corpus}
\usage{
tokenize(
  x,
  tokenizer,
  ...,
  hash_func = hash_string,
  minhash_func = NULL,
  keep_tokens = FALSE,
  keep_text = TRUE
)
}
\arguments{
\item{x}{A \code{\link{TextReuseTextDocument}} or
\code{\link{TextReuseCorpus}}.}

\item{tokenizer}{A function to split the text into tokens. See
\code{\link{tokenizers}}.}

\item{...}{Arguments passed on to the \code{tokenizer}.}

\item{hash_func}{A function to hash the tokens. See
\code{\link{hash_string}}.}

\item{minhash_func}{A function to create minhash signatures. See
\code{\link{minhash_generator}}.}

\item{keep_tokens}{Should the tokens be saved in the document that is
returned or discarded?}

\item{keep_text}{Should the text be saved in the document that is returned or
discarded?}
}
\value{
The modified \code{\link{TextReuseTextDocument}} or
  \code{\link{TextReuseCorpus}}.
}
\description{
Given a \code{\link{TextReuseTextDocument}} or a
\code{\link{TextReuseCorpus}}, this function recomputes the tokens and hashes
with the functions specified. Optionally, it can also recompute the minhash signatures.
}
\examples{
dir <- system.file("extdata/legal", package = "textreuse")
corpus <- TextReuseCorpus(dir = dir, tokenizer = NULL)
corpus <- tokenize(corpus, tokenize_ngrams)
head(tokens(corpus[[1]]))
}
